"""sort type enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from enum import auto, unique
from typing import List

from pygoodle.enum import AutoLowerName


@unique
class SortType(AutoLowerName):
    NAME = auto()
    SIZE = auto()
    DOWNLOADED = auto()
    TYPE = auto()

    @classmethod
    def all_values(cls) -> List[str]:
        return [e.value for e in SortType]



"""formatting utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import re
from pathlib import Path
from typing import Any, List, Optional

import humanize


class Format:

    @classmethod
    def bold(cls, output: Any) -> str:
        return f'[bold]{output}[/bold]'

    @classmethod
    def bool(cls, output: bool) -> str:
        if output:
            return Format.green(output)
        else:
            return Format.red(output)

    @classmethod
    def cyan(cls, output: Any) -> str:
        return f'[cyan]{output}[/cyan]'

    @classmethod
    def default(cls, output: Any) -> str:
        return f'[default]{output}[/default]'

    @classmethod
    def blue(cls, output: Any) -> str:
        return f'[blue]{output}[/blue]'

    @classmethod
    def green(cls, output: Any) -> str:
        return f'[green]{output}[/green]'

    @classmethod
    def red(cls, output: Any) -> str:
        return f'[red]{output}[/red]'

    @classmethod
    def magenta(cls, output: Any) -> str:
        return f'[magenta]{output}[/magenta]'

    @classmethod
    def yellow(cls, output: Any) -> str:
        return f'[yellow]{output}[/yellow]'

    @classmethod
    def escape(cls, output: Any) -> str:
        import rich.markup as markup
        return markup.escape(str(output))

    @classmethod
    def size(cls, size: int) -> str:
        parts = humanize.naturalsize(size).split()
        assert len(parts) == 2
        number = Format.bold(parts[0])
        unit = parts[1]
        return Format.green(f'{number} {unit}')

    @classmethod
    def underline(cls, output: Any) -> str:
        return f'[underline]{output}[/underline]'

    @classmethod
    def separator(cls, message: Any, character: str) -> str:
        sep = character * len(str(message))
        return f'[default bold]{sep}[/default bold]'

    @classmethod
    def h1(cls, message: Any, newline: bool = True) -> str:
        output = '\n' if newline else ''
        output = f'{output}[default bold]{message}[/default bold]'
        sep = Format.separator(message, '=')
        return f'{output}\n{sep}'

    @classmethod
    def h2(cls, message: Any, newline: bool = True) -> str:
        output = '\n' if newline else ''
        output = f'{output}[default bold]{message}[/default bold]'
        sep = Format.separator(message, '-')
        return f'{output}\n{sep}'

    @classmethod
    def h3(cls, message: Any, newline: bool = True) -> str:
        output = '\n' if newline else ''
        return f'{output}[default bold underline]# {message}[/default bold underline]'

    @classmethod
    def h4(cls, message: Any, newline: bool = True) -> str:
        output = '\n' if newline else ''
        return f'{output}[default bold underline]## {message}[/default bold underline]'

    @classmethod
    def h5(cls, message: Any, newline: bool = True) -> str:
        output = '\n' if newline else ''
        return f'{output}[default bold underline]### {message}[/default bold underline]'

    @classmethod
    def gnu_size(cls, size: int) -> str:
        return humanize.naturalsize(size, gnu=True)

    @classmethod
    def path(cls, path: Path, relative_to: Optional[Path] = None) -> str:
        import pygoodle.filesystem as fs
        if relative_to is not None and fs.is_relative_to(path, relative_to):
            path = path.relative_to(relative_to)
        return Format.cyan(path)

    @classmethod
    def symlink(cls, symlink: Path, relative_to: Optional[Path] = None, source: Optional[Path] = None) -> str:
        """Return formatted string for yaml symlink

        :param Path symlink: Yaml symlink
        :param Optional[Path] relative_to: Path to format symlink relative to
        :param Optional[Path] source: Manually specify source
        :return: Formatted string for yaml symlink
        """

        target_output = Format.path(symlink, relative_to=relative_to)
        if source is None:
            source_output = Format.path(symlink.resolve(), relative_to=relative_to)
        else:
            source_output = Format.path(source, relative_to=relative_to)
        return f'{target_output} -> {source_output}'

    @classmethod
    def get_lines(cls, path: Path) -> List[str]:
        contents = path.read_text().splitlines()
        lines = [line.strip() for line in contents]
        return lines

    @classmethod
    def list_from_string(cls, text: str, sep: Optional[str] = None) -> List[str]:
        return text.split(sep=sep)

    @classmethod
    def clean_escape_sequences(cls, string: str) -> str:
        reaesc = re.compile(r'\x1b[^m]*m')
        return reaesc.sub('', string)

    @classmethod
    def remove_prefix(cls, text: str, prefix: str) -> str:
        """Remove prefix from string

        :param str text: Text to remove prefix from
        :param str prefix: Prefix to remove
        :return: Text with prefix removed if present
        """

        if text.startswith(prefix):
            return text[len(prefix):]
        return text

    @classmethod
    def remove_suffix(cls, text: str, suffix: str) -> str:
        if text.endswith(suffix):
            return text[:len(suffix)]
        return text

    class Git:

        @classmethod
        def ref(cls, output: Any) -> str:
            return Format.magenta(output)

        @classmethod
        def remote(cls, output: Any) -> str:
            return Format.yellow(output)

        @classmethod
        def upstream(cls, output: Any) -> str:
            return Format.cyan(output)

        @classmethod
        def url(cls, output: any) -> str:
            return Format.cyan(output)



"""Misc git utils"""

from pathlib import Path
from subprocess import CompletedProcess
from typing import Dict, List, Optional

import pygoodle.command as cmd
import pygoodle.filesystem as fs
from pygoodle.format import Format

from .constants import HEAD, ORIGIN
from .process_output import ProcessOutput


class GitOnline:

    @classmethod
    def pull(cls, path: Path, remote: Optional[str] = None, branch: Optional[str] = None,
             rebase: bool = False, prune: bool = False, tags: bool = False,
             jobs: Optional[int] = None, no_edit: bool = False, autostash: bool = False,
             depth: Optional[int] = None, fetch_all: bool = False) -> CompletedProcess:
        """Pull upstream changes"""

        refspec = None
        if branch is not None:
            remote = ORIGIN if remote is None else remote
            refspec = f'refs/heads/{branch}:refs/remotes/{remote}/heads/{branch}'

        remote = '' if remote is None else remote
        refspec = '' if refspec is None else refspec

        args = ''
        if rebase:
            args += ' --rebase '
        if prune:
            args += ' --prune '
        if tags:
            args += ' --tags '
        if no_edit:
            args += ' --no-edit '
        if autostash:
            args += ' --autostash '
        if jobs is not None:
            args += f' --jobs={jobs} '
        if depth is not None:
            args += f' --depth={depth} '
        if fetch_all:
            args += f' --all '
        return cmd.run(f'git pull {args} {remote} {refspec}', cwd=path)

    @classmethod
    def pull_lfs(cls, path: Path) -> CompletedProcess:
        """Pull lfs files"""

        return cmd.run('git lfs pull', cwd=path)

    # See: https://github.blog/2020-12-21-get-up-to-speed-with-partial-clone-and-shallow-clone/
    @classmethod
    def clone(cls, path: Path, url: str, depth: Optional[int] = None, branch: Optional[str] = None,
              tag: Optional[str] = None, jobs: Optional[int] = None, single_branch: bool = False,
              blobless: bool = False, treeless: bool = False, origin: Optional[str] = None) -> CompletedProcess:
        if path.is_dir():
            if fs.has_contents(path):
                raise Exception(f'Existing directory at clone path {path}')
            fs.remove_dir(path)

        args = ''
        if branch is not None:
            args += f' --branch {branch} '
        elif tag is not None:
            args += f' --branch {tag} '

        if single_branch:
            args += ' --single-branch '
        if jobs is not None:
            args += f' --jobs {jobs} '
        if depth is not None:
            args += f' --depth {depth} '
        if origin is not None:
            args += f' --origin {origin} '

        assert not (blobless and treeless)
        if blobless:
            args += ' --filter=blob:none '
        elif treeless:
            args += ' --filter=tree:0 '

        return cmd.run(f'git clone {args} {url} {path}')

    @classmethod
    def push(cls, path: Path, remote: Optional[str] = None, local_branch: Optional[str] = None,
             remote_branch: Optional[str] = None, force: bool = False, set_upstream: bool = False) -> CompletedProcess:
        refspec = None
        if remote_branch is not None:
            remote = ORIGIN if remote is None else remote
            local_branch = HEAD if local_branch is None else local_branch
            refspec = f'refs/heads/{local_branch}:refs/heads/{remote_branch}'
        elif local_branch is not None:
            remote = ORIGIN if remote is None else remote
            remote_branch = local_branch
            refspec = f'refs/heads/{local_branch}:refs/heads/{remote_branch}'

        remote = '' if remote is None else remote
        refspec = '' if refspec is None else refspec

        args = ''
        if force:
            args += ' --force '
        if set_upstream:
            args += ' --set-upstream '

        return cmd.run(f'git push {args} {remote} {refspec}', cwd=path)

    @classmethod
    def fetch(cls, path: Path, prune: bool = False, prune_tags: bool = False, tags: bool = False,
              depth: Optional[int] = None, remote: Optional[str] = None, branch: Optional[str] = None,
              unshallow: bool = False, jobs: Optional[int] = None, fetch_all: bool = False,
              print_output: bool = True) -> CompletedProcess:

        refspec = None
        if branch is not None:
            remote = ORIGIN if remote is None else remote
            refspec = f'refs/heads/{branch}:refs/remotes/{remote}/heads/{branch}'

        remote = '' if remote is None else remote
        refspec = '' if refspec is None else refspec

        args = ''
        if prune:
            args += ' --prune '
        if prune_tags:
            args += ' --prune-tags '
        if tags:
            args += ' --tags '
        if depth is not None:
            args += f' --depth {depth}'
        if unshallow:
            args += ' --unshallow '
        if jobs is not None:
            args += ' --jobs '
        if fetch_all:
            args += ' --all '

        return cmd.run(f"git fetch {args} {remote} {refspec}", cwd=path, print_output=print_output)

    @classmethod
    def delete_remote_tag(cls, path: Path, tag: str, remote: Optional[str] = None,
                          force: bool = False) -> CompletedProcess:
        refspec = f':refs/tags/{tag}'
        remote = ORIGIN if remote is None else remote
        args = ''
        if force:
            args += ' --force '
        return cmd.run(f'git push {remote} {args} {refspec}', cwd=path)

    @classmethod
    def delete_remote_branch(cls, path: Path, branch: str, remote: str = ORIGIN,
                             force: bool = False) -> CompletedProcess:
        refspec = f':refs/heads/{branch}'
        remote = ORIGIN if remote is None else remote
        args = ''
        if force:
            args += ' --force '
        return cmd.run(f'git push {remote} {args} {refspec}', cwd=path)

    @classmethod
    def branch_exists_at_remote_url(cls, url: str, branch: str) -> bool:
        output = cmd.get_stdout(f'git ls-remote --heads {url} {branch}')
        if output is None:
            # TODO: Should this return None?
            return False
        return bool(output)

    @classmethod
    def branch_exists_on_remote(cls, path: Path, branch: str, remote: str = ORIGIN) -> bool:
        output = cmd.get_stdout(f'git ls-remote --heads {remote} {branch}', cwd=path)
        if output is None:
            # TODO: Should this return None?
            return False
        return bool(output)

    @classmethod
    def get_default_branch(cls, url: str) -> Optional[str]:
        """Get default branch from remote repo"""

        command = f'git ls-remote --symref {url} {HEAD}'
        output = cmd.get_stdout(command)
        if output is None:
            return None
        output_list = output.split()
        branch = [Format.remove_prefix(chunk, 'refs/heads/')
                  for chunk in output_list if chunk.startswith('refs/heads/')]
        return branch[0]

    @classmethod
    def submodule_update(cls, path: Path, init: bool = False, depth: Optional[int] = None, single_branch: bool = False,
                         jobs: Optional[int] = None, recursive: bool = False, remote: bool = False,
                         no_fetch: bool = False, checkout: bool = False, rebase: bool = False, merge: bool = False,
                         paths: Optional[List[Path]] = None) -> CompletedProcess:
        args = ''
        if init:
            args += ' --init '
        if single_branch:
            args += ' --single-branch '
        if jobs is not None:
            args += f' --jobs {jobs} '
        if depth is not None:
            args += f' --depth {depth} '
        if recursive is not None:
            args += ' --recursive '
        if remote:
            args += ' --remote '
        if no_fetch:
            args += ' --no-fetch '

        # TODO: Validate that at most one of these is True
        if checkout:
            args += ' --checkout '
        if merge:
            args += ' --merge '
        if rebase:
            args += ' --rebase '

        if paths is not None and paths:
            paths = ' '.join([str(p) for p in paths])
        else:
            paths = ''
        return cmd.run(f'git submodule update {args} {paths}', cwd=path)

    @classmethod
    def get_remote_tag_sha(cls, path: Path, name: str, remote: str = ORIGIN) -> Optional[str]:
        tags = GitOnline.get_remote_tags_info(path, remote=remote)
        sha = [sha for t, sha in tags.items() if t == name]
        return sha[0] if sha else None

    @classmethod
    def get_remote_tags_info(cls, path: Path = Path.cwd(), remote: str = ORIGIN) -> Dict[str, str]:
        output = cmd.get_stdout(f"git ls-remote --tags {remote}", cwd=path)
        if output is None:
            return {}
        return ProcessOutput.tag_shas(output)

    @classmethod
    def get_remote_tag(cls, path: Path, name: str, remote: str = ORIGIN) -> Optional[str]:
        tags = GitOnline.get_remote_tags_info(path, remote=remote)
        tag = [t for t, sha in tags.items() if t == name]
        return tag[0] if tag else None

    @classmethod
    def get_remote_branches_info(cls, path: Path = Path.cwd(), remote: str = ORIGIN) -> Dict[str, str]:
        output = cmd.get_stdout(f'git ls-remote --heads {remote}', cwd=path)
        if output is None:
            return {}
        return ProcessOutput.branch_shas(output)



"""Base Git utility class

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import List, Optional, Union, TYPE_CHECKING

from pygoodle.console import CONSOLE
# from pygoodle.git.decorators import not_detached
from pygoodle.format import Format
from pygoodle.git.constants import GitConfig, HEAD, ORIGIN
from pygoodle.git.decorators import error_msg
from pygoodle.git.log import GIT_LOG
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .branch.local_branch import LocalBranch
from .branch.remote_branch import RemoteBranch
from .branch.tracking_branch import TrackingBranch
from .protocol import Protocol
from .ref import Ref
from .remote import Remote
from .tag.local_tag import LocalTag
from .tag.remote_tag import RemoteTag

if TYPE_CHECKING:
    from .diff import Diff
    from .submodule import Submodule
    from .factory import AllBranches


class Repo:
    """Class encapsulating base git utilities

    :ivar Path path: Absolute path to repo
    :ivar Remote default_remote: Default remote
    """

    def __init__(self, path: Path, default_remote: Optional[str] = None, url: Optional[str] = None,
                 protocol: Protocol = Protocol.SSH):
        """LocalRepo __init__

        :param Path path: Absolute path to repo
        :param str default_remote: Default remote name
        """

        default_remote = ORIGIN if default_remote is None else default_remote

        self.path: Path = path
        self.default_remote: Remote = Remote(self.path, default_remote)
        self.url: Optional[str] = url
        self.protocol: Protocol = protocol

    @property
    def git_dir(self) -> Optional[Path]:
        return GitOffline.git_dir(self.path)

    @error_msg('Failed to clone repo')
    def clone(self, path: Path, url: str, depth: Optional[int] = None, branch: Optional[str] = None,
              jobs: Optional[int] = None, origin: Optional[str] = None) -> 'Repo':
        CONSOLE.stdout(' - Clone repo')
        GitOnline.clone(path, url=url, depth=depth, branch=branch, jobs=jobs, origin=origin)
        return Repo(path)

    @property
    def untracked_files(self) -> List[Path]:
        return GitOffline.get_untracked_files(self.path)

    @property
    def has_untracked_files(self) -> bool:
        return GitOffline.has_untracked_files(self.path)

    @property
    def is_dirty(self) -> bool:
        return GitOffline.is_dirty(self.path)

    @property
    def is_detached(self) -> bool:
        return GitOffline.is_detached(self.path)

    @property
    def is_shallow(self) -> bool:
        return GitOffline.is_shallow_repo(self.path)

    @property
    def is_rebase_in_progress(self) -> bool:
        return GitOffline.is_rebase_in_progress(self.path)

    def get_remotes(self) -> List[Remote]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_remotes(self.path)

    def get_submodules(self) -> List['Submodule']:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_submodules(self.path)

    def get_tracking_branches(self) -> List[TrackingBranch]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_tracking_branches(self.path)

    def get_local_branches(self) -> List[LocalBranch]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_local_branches(self.path)

    def get_remote_branches(self, online: bool = False) -> List[RemoteBranch]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_all_remote_branches(self.path, online=online)

    def get_all_branches(self, online: bool = False) -> 'AllBranches':
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_all_branches(self.path, online=online)

    def has_local_branch(self, branch: str) -> bool:
        local_branch = self.get_local_branch(branch)
        return local_branch is not None

    def has_remote_branch(self, branch: str, remote: Optional[str] = None,
                          url: Optional[str] = None, online: bool = False) -> bool:
        remote_branch = self.get_remote_branch(branch, remote=remote, url=url, online=online)
        return remote_branch is not None

    def has_tracking_branch(self, branch: str, remote: Optional[str] = None) -> bool:
        tracking_branch = self.get_tracking_branch(branch, remote=remote)
        return tracking_branch is not None

    def has_submodule(self, submodule_path: Path) -> bool:
        submodule = self.get_submodule(submodule_path)
        return submodule is not None

    def get_submodule(self, submodule_path: Path) -> Optional['Submodule']:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_submodule(self.path, submodule_path)

    def get_local_branch(self, branch: str) -> Optional[LocalBranch]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_local_branch(self.path, branch)

    def get_remote_branch(self, branch: str, remote: Optional[str] = None,
                          url: Optional[str] = None, online: bool = False) -> Optional[RemoteBranch]:
        remote = ORIGIN if remote is None else remote
        from pygoodle.git.model.factory import GitFactory
        if online or url is not None:
            return GitFactory.get_remote_branch_online(self.path, branch=branch, remote=remote, url=url)
        return GitFactory.get_remote_branch_offline(self.path, branch=branch, remote=remote)

    def get_local_tags(self) -> List[LocalTag]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_local_tags(self.path)

    def get_local_tag(self, tag: str) -> Optional[LocalTag]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_local_tag(self.path, tag)

    def has_local_tag(self, tag: str) -> bool:
        local_tag = self.get_local_tag(tag)
        return local_tag is not None

    def has_remote_tag(self, tag: str, remote: Optional[str] = None,
                       url: Optional[str] = None) -> bool:
        remote_tag = self.get_remote_tag(tag, remote=remote, url=url)
        return remote_tag is not None

    def get_remote_tag(self, tag: str, remote: Optional[str] = None,
                       url: Optional[str] = None) -> Optional[RemoteTag]:
        remote = ORIGIN if remote is None else remote
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_remote_tag(self.path, tag, remote, url=url)

    def get_remote_tags(self, remote: Optional[str] = None,
                        url: Optional[str] = None) -> List[RemoteTag]:
        remote = ORIGIN if remote is None else remote
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_remote_tags(self.path, remote, url=url)

    def get_remote(self, remote: Optional[str] = None, fetch_url: Optional[str] = None,
                   push_url: Optional[str] = None) -> Optional[Remote]:
        from pygoodle.git.model.factory import GitFactory
        remotes = GitFactory.get_remotes(self.path)
        if remote is not None:
            remotes = [r for r in remotes if r.name == remote]
        if fetch_url is not None:
            remotes = [r for r in remotes if r.fetch_url == fetch_url]
        if push_url is not None:
            remotes = [r for r in remotes if r.push_url == push_url]
        return remotes[0] if remotes else None

    def has_remote(self, remote: Optional[str] = None, fetch_url: Optional[str] = None,
                   push_url: Optional[str] = None) -> bool:
        remote = self.get_remote(remote=remote, fetch_url=fetch_url, push_url=push_url)
        return remote is not None

    def get_tracking_branch(self, branch: str, remote: Optional[str] = None) -> Optional[TrackingBranch]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_tracking_branch(self.path, branch, remote=remote)

    def get_diff(self) -> 'Diff':
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_diff(self.path)

    @property
    def exists(self) -> bool:
        return GitOffline.is_repo_cloned(self.path)

    def checkout(self, ref: str, track: bool = False) -> None:
        if self.is_dirty:
            CONSOLE.stdout(' - Dirty repo. Please stash, commit, or discard your changes')
            self.status(verbose=True)
            return
        GitOffline.checkout(self.path, ref=ref, track=track)

    def is_valid(self, allow_missing: bool = True) -> bool:
        """Validate repo state

        :param bool allow_missing: Whether to allow validation to succeed with missing repo
        :return: True, if repo not dirty or doesn't exist on disk
        """

        if not self.exists:
            return allow_missing

        if self.is_dirty or self.is_rebase_in_progress or self.has_untracked_files:
            return False

        submodules = self.get_submodules()
        if not submodules:
            return True
        return all([s.is_valid(allow_missing=allow_missing) for s in submodules])

    def remote(self, name: str) -> Optional[Remote]:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.get_remote(self.path, name)

    @property
    def current_timestamp(self) -> str:
        return GitOffline.current_timestamp(self.path)

    @property
    def current_branch(self) -> str:
        return GitOffline.current_branch(self.path)

    def sha(self, ref: Optional[str] = None, short: bool = False) -> str:
        if ref is None:
            return GitOffline.current_head_commit_sha(self.path, short=short)
        return GitOffline.get_sha(self.path, ref=ref, short=short)

    @error_msg('Failed to abort rebase')
    def abort_rebase(self) -> None:
        if not self.is_rebase_in_progress:
            return
        CONSOLE.stdout(' - Abort rebase in progress')
        GitOffline.abort_rebase(self.path)

    @error_msg('Failed to add files to git index')
    def add(self, files: Union[Path, str, List[str], List[Path]]) -> None:
        if isinstance(files, str):
            files = [files]
        elif isinstance(files, Path):
            files = [str(files)]
        elif isinstance(files, list):
            files = [str(f) for f in files]
        else:
            raise Exception('Wrong type for files parameter')

        CONSOLE.stdout(' - Add files to git index')
        GitOffline.add(self.path, files=files)

    @error_msg('Failed to commit current changes')
    def commit(self, message: str) -> None:
        CONSOLE.stdout(' - Commit current changes')
        GitOffline.commit(self.path, message=message)

    def clean(self, untracked_directories: bool = False, force: bool = False,
              ignored: bool = False, untracked_files: bool = False) -> None:
        """Discard changes for repo

        :param bool untracked_directories: ``d`` Remove untracked directories in addition to untracked files
        :param bool force: ``f`` Delete directories with .git sub directory or file
        :param bool ignored: ``X`` Remove only files ignored by git
        :param bool untracked_files: ``x`` Remove all untracked files
        """

        CONSOLE.stdout(' - Clean repo')
        GitOffline.clean(self.path, untracked_directories=untracked_directories,
                         force=force, ignored=ignored, untracked_files=untracked_files)

    @error_msg('Failed to pull git lfs files')
    def pull_lfs(self) -> None:
        CONSOLE.stdout(' - Pull git lfs files')
        GitOnline.pull_lfs(self.path)

    @error_msg('Failed to reset repo')
    def reset(self, ref: Union[Ref, str] = HEAD, hard: bool = False) -> None:
        if isinstance(ref, TrackingBranch):
            ref = ref.upstream_branch.short_ref
        elif isinstance(ref, Ref):
            ref = ref.short_ref
        CONSOLE.stdout(f' - Reset repo to {Format.Git.ref(ref)}')
        GitOffline.reset(self.path, ref=ref, hard=hard)

    @error_msg('Failed to stash current changes')
    def stash(self) -> None:
        if not self.is_dirty:
            CONSOLE.stdout(' - No changes to stash')
            return
        CONSOLE.stdout(' - Stash current changes')
        GitOffline.stash(self.path)

    def status(self, verbose: bool = False) -> None:
        GitOffline.status(self.path, verbose=verbose)

    @error_msg('Failed to update local git config')
    def update_git_config(self, config: GitConfig) -> None:
        """Update custom git config

        :param GitConfig config: Custom git config
        """

        CONSOLE.stdout(" - Update local git config")
        for key, value in config.items():
            GitOffline.git_config_unset_all_local(self.path, key)
            GitOffline.git_config_add_local(self.path, key, value)

    @error_msg('Failed to update git lfs hooks')
    def install_lfs_hooks(self, local: bool = False) -> None:
        CONSOLE.stdout(' - Update git lfs hooks')
        GitOffline.install_lfs_hooks(self.path, local=local)

    @error_msg('Failed to uninstall git lfs hooks and filters')
    def uninstall_lfs(self) -> None:
        CONSOLE.stdout(' - Uninstall git lfs hooks')
        GitOffline.uninstall_lfs_hooks(self.path)
        CONSOLE.stdout(' - Uninstall git lfs filters')
        GitOffline.uninstall_lfs_filters(self.path)

    @error_msg('Failed to reset timestamp')
    def reset_timestamp(self, timestamp: str, ref: Ref, author: Optional[str] = None) -> None:
        CONSOLE.stdout(' - Reset timestamp')
        GitOffline.reset_timestamp(self.path, timestamp=timestamp, ref=ref.short_ref, author=author)

    def submodule_add(self, url: str, branch: Optional[str] = None, force: bool = False,
                      name: Optional[str] = None, reference: Optional[str] = None, depth: Optional[int] = None,
                      submodule_path: Optional[Path] = None) -> None:
        GitOffline.submodule_add(self.path, url, branch=branch, force=force, name=name,
                                 reference=reference, depth=depth, submodule_path=submodule_path)

    @error_msg('Failed to update submodules')
    def submodule_update(self, init: bool = False, depth: Optional[int] = None, single_branch: bool = False,
                         jobs: Optional[int] = None, recursive: bool = False, remote: bool = False,
                         checkout: bool = False, rebase: bool = False, merge: bool = False,
                         paths: Optional[List[Path]] = None) -> None:
        CONSOLE.stdout(' - Update submodules')
        GitOnline.submodule_update(self.path, init=init, depth=depth, single_branch=single_branch,
                                   jobs=jobs, recursive=recursive, remote=remote, checkout=checkout,
                                   merge=merge, rebase=rebase, paths=paths)

    @error_msg('Failed to deinit submodules')
    def submodule_deinit(self, force: bool = False, paths: Optional[List[Path]] = None) -> None:
        CONSOLE.stdout(' - Deinit submodules')
        GitOffline.submodule_deinit(self.path, force=force, paths=paths)

    @error_msg('Failed to init submodules')
    def submodule_init(self, paths: Optional[List[Path]] = None) -> None:
        CONSOLE.stdout(' - Init submodules')
        GitOffline.submodule_init(self.path, paths=paths)

    @error_msg('Failed to sync submodules')
    def submodule_sync(self, recursive: bool = False, paths: Optional[List[Path]] = None) -> None:
        CONSOLE.stdout(' - Sync submodules')
        GitOffline.submodule_sync(self.path, recursive=recursive, paths=paths)

    def print_local_branches(self) -> None:
        """Print local git branches"""

        for branch in self.get_local_branches():
            if branch.name == self.current_branch:
                CONSOLE.stdout(f'* {Format.green(branch.name)}')
            else:
                CONSOLE.stdout(branch)

    def print_validation(self, allow_missing: bool = False) -> None:
        """Print validation message"""

        if not self.exists or self.is_valid(allow_missing=allow_missing):
            return
        self.status()
        CONSOLE.stdout(f'Dirty repo. Please stash, commit, or discard your changes')

    def groom(self, untracked_directories: bool = True, force: bool = True,
              ignored: bool = False, untracked_files: bool = True) -> None:
        self.clean(untracked_directories=untracked_directories,
                   force=force, ignored=ignored, untracked_files=untracked_files)
        self.reset(hard=True)
        if self.is_rebase_in_progress:
            self.abort_rebase()

    @property
    def formatted_ref(self) -> str:
        """Formatted project repo ref"""

        if self.is_detached:
            return Format.Git.ref(Format.escape(f'[HEAD @ {self.sha()}]'))

        current_branch_output = Format.Git.ref(Format.escape(f'[{self.current_branch}]'))

        local_commits_count = GitOffline.new_commits_count(self.path)
        no_local_commits = local_commits_count == 0
        # TODO: Specify correct remote
        upstream_commits_count = GitOffline.new_commits_count(self.path, upstream=True)
        no_upstream_commits = upstream_commits_count == 0

        if no_local_commits and no_upstream_commits:
            return current_branch_output

        local_commits_output = Format.yellow(f'+{local_commits_count}')
        upstream_commits_output = Format.red(f'-{upstream_commits_count}')
        return f'{current_branch_output}({local_commits_output}/{upstream_commits_output})'

    def print_remote_branches(self) -> None:
        """Print remote git branches"""

        # FIXME: Update this to work
        # Need to get all local, remote, and tracking branches and print them
        # for remote in self.remotes:
        #     for branch in remote.branches:
        #     if ' -> ' in branch:
        #         components = branch.split(' -> ')
        #         local_branch = components[0]
        #         remote_branch = components[1]
        #         CONSOLE.stdout(f"  {Format.red(local_branch)} -> {remote_branch}")
        #     else:
        #         CONSOLE.stdout(Format.red(branch))

    def fetch(self, prune: bool = False, prune_tags: bool = False, tags: bool = False,
              depth: Optional[int] = None, remote: Optional[str] = None, branch: Optional[str] = None,
              unshallow: bool = False, jobs: Optional[int] = None, fetch_all: bool = False, check: bool = True,
              print_output: bool = True):
        # FIXME: Consolidate this implementation with the one for Remotes
        CONSOLE.stdout(f' - Fetch repo')
        try:
            GitOnline.fetch(self.path, remote=remote, prune=prune, prune_tags=prune_tags, tags=tags, depth=depth,
                            branch=branch, unshallow=unshallow, jobs=jobs, fetch_all=fetch_all, print_output=print_output)
        except Exception:  # noqa
            message = f'Failed to fetch repo'
            if check:
                GIT_LOG.error(message)
                raise
            CONSOLE.stdout(message)

    @error_msg('Failed to pull')
    # @not_detached
    def pull(self, remote: Optional[str] = None, branch: Optional[str] = None,
             rebase: bool = False, prune: bool = False, tags: bool = False,
             jobs: Optional[int] = None, no_edit: bool = False, autostash: bool = False,
             depth: Optional[int] = None) -> None:
        # TODO: Check if detached
        message = f' - Pull'
        if rebase:
            message += ' with rebase'
        CONSOLE.stdout(message)
        GitOnline.pull(self.path, remote=remote, branch=branch, rebase=rebase, prune=prune, tags=tags,
                       jobs=jobs, no_edit=no_edit, autostash=autostash, depth=depth)

    # @error_msg('Failed to push')
    # @not_detached
    def push(self, remote: Optional[str] = None, local_branch: Optional[str] = None,
             remote_branch: Optional[str] = None, force: bool = False, set_upstream: bool = False) -> None:
        # TODO: Check if detached
        CONSOLE.stdout(' - Push current branch')
        GitOnline.push(self.path,
                       remote=remote,
                       local_branch=local_branch,
                       remote_branch=remote_branch,
                       force=force,
                       set_upstream=set_upstream)



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from typing import Any, Dict, List

from .argument import Argument


class ArgumentGroup:

    def __init__(self, args: List[Argument], title: str, **kwargs):
        self.title: str = title
        self.args: List[Argument] = args
        self.options: Dict[str, Any] = kwargs



"""git remote

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import List, Optional

from pygoodle.console import CONSOLE
from pygoodle.format import Format
# from pygoodle.git.decorators import error_msg
from pygoodle.git.log import GIT_LOG
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .branch.remote_branch import RemoteBranch


class Remote:
    """Class encapsulating git ref

    :ivar Path path: Path to git repo
    :ivar str name: Branch
    :ivar str fetch_url: Fetch url
    :ivar str push_url: Push url
    """

    def __init__(self, path: Path, name: str):
        """GitRemote __init__

        :param Path path: Path to git repo
        :param str name: Branch
        """

        self.name: str = name
        self.path: Path = path

    def __eq__(self, other) -> bool:
        if isinstance(other, Remote):
            return self.path == other.path and self.name == other.name
        return False

    def __lt__(self, other: 'Remote') -> bool:
        return self.name < other.name

    @property
    def fetch_url(self) -> str:
        return GitOffline.get_remote_fetch_url(self.path, self.name)

    @property
    def push_url(self) -> str:
        return GitOffline.get_remote_push_url(self.path, self.name)

    def branches(self, url: Optional[str] = None, online: bool = False) -> List[RemoteBranch]:
        from pygoodle.git.model.factory import GitFactory
        if online or url is not None:
            return GitFactory.get_remote_branches_online(self.path, remote=self.name, url=url)
        return GitFactory.get_remote_branches_offline(self.path, self.name)

    # @error_msg('Failed to create remote')
    def create(self, url: str, fetch: bool = False, tags: bool = False) -> None:
        if self.exists:
            CONSOLE.stdout(f' - Remote {Format.Git.remote(self.name)} already exists')
            return
        CONSOLE.stdout(f' - Create remote {Format.Git.remote(self.name)}')
        GitOffline.create_remote(self.path, name=self.name, url=url, fetch=fetch, tags=tags)

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_remote(self.path, remote=self.name)

    def default_branch(self, url: str) -> Optional[RemoteBranch]:
        if GitOffline.is_repo_cloned(self.path):
            default_branch = GitOffline.get_default_branch(self.path, self.name)
            if default_branch is not None:
                return RemoteBranch(self.path, default_branch, self.name)
        default_branch = GitOnline.get_default_branch(url)
        if default_branch is None:
            return None
        git_dir = GitOffline.git_dir(self.path)
        if git_dir is not None and git_dir.is_dir():
            GitOffline.save_default_branch(git_dir, self.name, default_branch)
        return RemoteBranch(self.path, default_branch, self.name)

    # @error_msg('Failed to rename remote')
    def rename(self, name: str) -> None:
        CONSOLE.stdout(f' - Rename remote {Format.Git.remote(self.name)} to {Format.Git.remote(name)}')
        GitOffline.rename_remote(self.path, old_name=self.name, new_name=name)
        self.name = name

    def fetch(self, prune: bool = False, prune_tags: bool = False, tags: bool = False,
              depth: Optional[int] = None, branch: Optional[str] = None, unshallow: bool = False,
              jobs: Optional[int] = None, fetch_all: bool = False, check: bool = True,
              print_output: bool = True) -> None:
        output = self.name
        if branch is not None:
            branch = branch
            output = f'{output} {branch}'
        CONSOLE.stdout(f'Fetch from {output}')
        try:
            GitOnline.fetch(self.path, remote=self.name, prune=prune, prune_tags=prune_tags, tags=tags, depth=depth,
                            branch=branch, unshallow=unshallow, jobs=jobs, fetch_all=fetch_all, print_output=print_output)
        except Exception:  # noqa
            message = f'Failed to fetch from {output}'
            if check:
                GIT_LOG.error(message)
                raise
            CONSOLE.stdout(f' - {message}')

    def print_branches(self) -> None:
        raise NotImplementedError

    def _compare_remote_url(self, remote: str, url: str) -> None:
        """Compare actual remote url to given url

        If URL's are different print error message and exit

        :param str remote: Remote name
        :param str url: URL to compare with remote's URL
        :raise ClowderGitError:
        """

        # FIXME: Implement
        # if url != self._remote_get_url(remote):
        #     actual_url = self._remote_get_url(remote)
        #     message = f"Remote {fmt.remote(remote)} already exists with a different url\n" \
        #               f"{fmt.url_string(actual_url)} should be {fmt.url_string(url)}"
        #     raise ClowderGitError(message)



"""git model factory

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import List, Optional, Tuple

from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .change import Change
from .diff import Diff
from .branch.local_branch import LocalBranch
from .branch.remote_branch import RemoteBranch
from .branch.tracking_branch import TrackingBranch
from .remote import Remote
from .submodule import Submodule
from .tag.local_tag import LocalTag
from .tag.remote_tag import RemoteTag


class AllBranches:

    def __init__(self, local_branches: List[LocalBranch], remote_branches: List[RemoteBranch],
                 tracking_branches: List[TrackingBranch]):

        local_branches = [b for b in local_branches if not b.is_tracking_branch]
        remote_branches = [b for b in remote_branches if not b.is_tracking_branch]
        self.local_branches: Tuple[LocalBranch, ...] = tuple(local_branches)
        self.remote_branches: Tuple[RemoteBranch, ...] = tuple(remote_branches)
        self.tracking_branches: Tuple[TrackingBranch, ...] = tuple(tracking_branches)


class GitFactory:

    @classmethod
    def get_diff(cls, path: Path) -> Diff:
        diff_info = GitOffline.get_diff_index_info(path)
        changes = []
        for change_type, changes_list in diff_info.items():
            for change_info in changes_list:
                change = Change(path,
                                file_path=Path(change_info['file_path']),
                                change_type=change_type,
                                old_permissions=change_info['old_permissions'],
                                new_permissions=change_info['new_permissions'],
                                old_sha=change_info['old_sha'],
                                new_sha=change_info['new_sha'])
                changes.append(change)
        return Diff(path, changes=changes)

    @classmethod
    def get_remotes(cls, path: Path) -> List[Remote]:
        remotes = GitOffline.get_remotes_info(path)
        remotes = [Remote(path, name) for name in remotes.keys()]
        return sorted(remotes)

    @classmethod
    def has_remote(cls, path: Path, remote: Optional[str] = None, fetch_url: Optional[str] = None,
                   push_url: Optional[str] = None) -> bool:
        remote = GitFactory.get_remote(path, remote=remote, fetch_url=fetch_url, push_url=push_url)
        return remote is not None

    @classmethod
    def get_remote(cls, path: Path, remote: Optional[str] = None, fetch_url: Optional[str] = None,
                   push_url: Optional[str] = None) -> Optional[Remote]:
        remotes = GitFactory.get_remotes(path)
        if remote is not None:
            remotes = [r for r in remotes if r.name == remote]
        if fetch_url is not None:
            remotes = [r for r in remotes if r.fetch_url == fetch_url]
        if push_url is not None:
            remotes = [r for r in remotes if r.push_url == push_url]
        return remotes[0] if remotes else None

    @classmethod
    def get_local_branches(cls, path: Path) -> List[LocalBranch]:
        branches = GitOffline.get_local_branches_info(path)
        branches = [LocalBranch(path, branch) for branch in branches]
        return sorted(branches)

    @classmethod
    def get_local_tags(cls, path: Path) -> List[LocalTag]:
        tags = GitOffline.get_local_tags_info(path)
        tags = [LocalTag(path, tag) for tag in tags]
        return sorted(tags)

    @classmethod
    def get_submodule(cls, path: Path, submodule_path: Path) -> Optional[Remote]:
        submodules = GitFactory.get_submodules(path)
        full_path = path / submodule_path
        submodule = [s for s in submodules if s.path == full_path]
        return submodule[0] if submodule else None

    @classmethod
    def get_submodules(cls, path: Path) -> List[Submodule]:
        submodules_info = GitOffline.get_submodules_info(path)
        submodules = []
        for key in submodules_info.keys():
            submodule_info = submodules_info[key]
            # TODO: Save url from config and gitmodules
            url = submodule_info['url']
            submodule_path = Path(key)
            branch = submodule_info['branch'] if 'branch' in submodule_info else None
            active = submodule_info['active'] if 'active' in submodule_info else None
            active = True if active == 'true' else False
            submodule_commit = GitOffline.get_submodule_commit(path, submodule_path)
            submodule = Submodule(path, submodule_path, url=url, commit=submodule_commit, branch=branch, active=active)
            submodules.append(submodule)
            recursive_submodules = GitFactory.get_submodules(submodule.path)
            if recursive_submodules:
                submodules += recursive_submodules
        return sorted(submodules)

    @classmethod
    def has_submodule(cls, path: Path, submodule_path: Path) -> bool:
        submodule = GitFactory.get_submodule(path, submodule_path)
        return submodule is not None

    @classmethod
    def get_tracking_branches(cls, path: Path) -> List[TrackingBranch]:
        branches = GitOffline.get_tracking_branches_info(path)
        tracking_branches = []
        for local_branch, info in branches.items():
            tracking_branch = TrackingBranch(path,
                                             local_branch=local_branch,
                                             upstream_branch=info['upstream_branch'],
                                             upstream_remote=info['upstream_remote'],
                                             push_branch=info['push_branch'],
                                             push_remote=info['push_remote'])
            tracking_branches.append(tracking_branch)
        return sorted(tracking_branches)

    @classmethod
    def get_all_branches(cls, path: Path, online: bool = False) -> AllBranches:
        local_branches = GitFactory.get_local_branches(path)
        remote_branches = GitFactory.get_all_remote_branches(path, online=online)
        tracking_branches = GitFactory.get_tracking_branches(path)
        return AllBranches(
            local_branches=local_branches,
            remote_branches=remote_branches,
            tracking_branches=tracking_branches
        )

    @classmethod
    def get_all_remote_branches(cls, path: Path, online: bool = False) -> List[RemoteBranch]:
        branches = []
        for remote in GitFactory.get_remotes(path):
            branches += remote.branches(online=online)
        return sorted(branches)

    @classmethod
    def get_remote_branches_offline(cls, path: Path, remote: str) -> List[RemoteBranch]:
        branches, default_branch = GitOffline.get_remote_branches_info(path, remote)
        branches = [RemoteBranch(path, branch, remote) for branch in branches]
        if default_branch is not None:
            branches.append(RemoteBranch(path, default_branch, remote, is_default=True))
        return sorted(branches)

    @classmethod
    def get_remote_branches_online(cls, path: Path, remote: str,
                                   url: Optional[str] = None) -> List[RemoteBranch]:
        if url is None:
            branches = GitOnline.get_remote_branches_info(path, remote=remote)
        else:
            branches = GitOnline.get_remote_branches_info(remote=url)
        branches = [RemoteBranch(path, branch, remote) for branch in branches]
        return sorted(branches)

    @classmethod
    def get_remote_tags(cls, path: Path, remote: str, url: Optional[str] = None) -> List[RemoteTag]:
        if url is None:
            tags = GitOnline.get_remote_tags_info(path, remote=remote)
        else:
            tags = GitOnline.get_remote_tags_info(remote=url)
        tags = [RemoteTag(path, tag, remote) for tag in tags]
        return sorted(tags)

    @classmethod
    def get_local_branch(cls, path: Path, branch: str) -> Optional[LocalBranch]:
        branches = GitFactory.get_local_branches(path)
        branch = [b for b in branches if b.name == branch]
        return branch[0] if branch else None

    @classmethod
    def has_local_branch(cls, path: Path, branch: str) -> bool:
        branch = GitFactory.get_local_branch(path, branch)
        return branch is not None

    @classmethod
    def get_remote_branch_offline(cls, path: Path, branch: str, remote: str) -> Optional[RemoteBranch]:
        branches = GitFactory.get_remote_branches_offline(path, remote)
        branch = [b for b in branches if b.name == branch]
        return branch[0] if branch else None

    @classmethod
    def has_remote_branch_offline(cls, path: Path, branch: str, remote: str) -> bool:
        branch = GitFactory.get_remote_branch_offline(path, branch, remote)
        return branch is not None

    @classmethod
    def has_remote_branch_online(cls, path: Path, branch: str, remote: str, url: Optional[str] = None) -> bool:
        branch = GitFactory.get_remote_branch_online(path, branch, remote=remote, url=url)
        return branch is not None

    @classmethod
    def get_remote_branch_online(cls, path: Path, branch: str, remote: str,
                                 url: Optional[str] = None) -> Optional[RemoteBranch]:
        branches = GitFactory.get_remote_branches_online(path, remote=remote, url=url)
        branch = [b for b in branches if b.name == branch]
        return branch[0] if branch else None

    @classmethod
    def get_tracking_branch(cls, path: Path, branch: str, remote: Optional[str] = None) -> Optional[TrackingBranch]:
        branches = GitFactory.get_tracking_branches(path)
        branch = [b for b in branches if b.name == branch]
        if remote is None:
            return branch[0] if branch else None
        branch = [b for b in branch if b.upstream_branch.remote.name == remote]
        return branch[0] if branch else None

    @classmethod
    def has_tracking_branch(cls, path: Path, branch: str) -> bool:
        branch = GitFactory.get_tracking_branch(path, branch)
        return branch is not None

    @classmethod
    def get_local_tag(cls, path: Path, tag: str) -> Optional[LocalTag]:
        tags = GitFactory.get_local_tags(path)
        tag = [t for t in tags if t.name == tag]
        return tag[0] if tag else None

    @classmethod
    def has_local_tag(cls, path: Path, tag: str) -> bool:
        tag = GitFactory.get_local_tag(path, tag)
        return tag is not None

    @classmethod
    def get_remote_tag(cls, path: Path, tag: str, remote: str, url: Optional[str] = None) -> Optional[RemoteTag]:
        tags = GitFactory.get_remote_tags(path, remote, url=url)
        tag = [t for t in tags if t.name == tag]
        return tag[0] if tag else None

    @classmethod
    def has_remote_tag(cls, path: Path, tag: str, remote: str, url: Optional[str] = None) -> bool:
        tag = GitFactory.get_remote_tag(path, tag=tag, remote=remote, url=url)
        return tag is not None



"""pygoodle.ftp module __init__

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .file_type import FileType
from .file_wrapper import FileWrapper
from .sort_type import SortType
from .ftp import FTP, CancelledDownloadError



"""ftp

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import ftplib
from pathlib import Path
from typing import Callable, List, Generator

from resource_pool import LazyPool

import pygoodle.filesystem as fs
from pygoodle.tasks import ProgressTask, ProgressTaskPool
from pygoodle.console import CONSOLE, disable_output

from .file_type import FileType
from .file_wrapper import FileWrapper


class CancelledDownloadError(Exception):
    pass


class FileInfo:

    def __init__(self, path: Path, size: int):
        self.path: Path = path
        self.name: str = path.name
        self.size: int = size

    def __lt__(self, other: 'FileInfo') -> bool:
        return str(self.path).lower() < str(other.path).lower()


class DirectoryInfo:

    def __init__(self, ftp: ftplib.FTP, directory: Path, recursive: bool = True):
        self.directory: Path = directory
        self.files: List[FileInfo] = []
        self._directories: List[Path] = []
        self.recursive = recursive

        self._load_directory(ftp, directory)

    def _load_directory(self, ftp: ftplib.FTP, directory: Path) -> None:
        results = [((directory / name).resolve(), facts)
                   for name, facts in ftp.mlsd(str(directory), facts=["type", "size"])]

        files = [FileInfo(name, int(facts['size'])) for name, facts in results if facts['type'] == 'file']
        self.files += files

        directories = [name for name, facts in results if facts['type'] == 'dir']
        self._directories += directories

        if not self.recursive:
            return

        for directory in directories:
            self._load_directory(ftp, directory)

    @property
    def size(self) -> int:
        if not self.recursive:
            raise Exception('Size can only be calculated when recursive')
        if not self.files:
            return 0
        return sum([f.size for f in self.files])

    @property
    def directories(self) -> List[Path]:
        directory_paths = [d for d in self._directories]
        return sorted(directory_paths, key=lambda path: str(path).lower())

    @property
    def file_paths(self) -> List[Path]:
        file_paths = [f.path for f in self.files]
        return sorted(file_paths, key=lambda path: str(path).lower())

    @property
    def all_paths(self) -> List[Path]:
        file_paths = [f.path for f in self.files]
        directory_paths = [d for d in self._directories]
        return sorted(file_paths + directory_paths, key=lambda path: str(path).lower())


class FTP:

    def __init__(self, host: str, user: str, password: str,
                 download_dir: Path, base_dir: Path = Path('/'), pool_size: int = 1):
        self.download_dir: Path = download_dir
        self.pool_size: int = pool_size
        self.base_dir = base_dir

        def make_ftp() -> ftplib.FTP:
            return ftplib.FTP(host=host, user=user, passwd=password)
        self._ftp_pool: LazyPool[ftplib.FTP] = LazyPool(factory=make_ftp, pool_size=pool_size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        while len(self._ftp_pool) > 0:
            with self._ftp_pool.get() as ftp:
                ftp.close()

    @disable_output
    def download(self, path: Path) -> Path:
        file_type = self._file_type(path)
        if file_type is FileType.FILE:
            files = [FileInfo(path, self._get_file_size(path))]
        elif file_type is FileType.DIR:
            with self.reserve_ftp() as ftp:
                directory_info = DirectoryInfo(ftp, path)
            self._make_download_dirs(path, directory_info)
            files = directory_info.files
        else:
            raise Exception('Unknown file type')

        download_path = self._get_download_path(path)

        class DownloadTaskPool(ProgressTaskPool):
            def after_tasks(self, tasks: List[ProgressTask]) -> None:
                super().after_tasks(tasks)
                if self.cancelled:
                    CONSOLE.stdout(f'Remove {download_path}', force=True)
                    fs.remove(download_path)

        ftp = self

        class DownloadTask(ProgressTask):
            def __init__(self, file: FileInfo):
                super().__init__(file.name, units='bytes', total=file.size)
                self._file: FileInfo = file

            def run(self) -> None:
                def callback(data):
                    if self.cancelled:
                        raise CancelledDownloadError(f'FTP download of {self._file.name} was cancelled')
                    self.progress.update_subtask(self._file.name, advance=len(data))

                ftp.download_file(self._file, callback)

        download_tasks = [DownloadTask(file) for file in files]
        pool = DownloadTaskPool(jobs=self.pool_size, title='Files', units='files')
        pool.run(download_tasks)
        CONSOLE.stdout(f'Downloaded {len(files)} files', force=True)
        return download_path

    def get_file_wrapper(self, path: Path) -> FileWrapper:
        file_type = self._file_type(path)
        if file_type is FileType.DIR:
            return FileWrapper(path.name, self._get_directory_size(path), file_type)
        elif file_type is FileType.FILE:
            return FileWrapper(path.name, self._get_file_size(path), file_type)

    def list_dir(self, directory: Path) -> List[Path]:
        if not directory.is_absolute():
            directory = self.base_dir / directory
        # FIXME: Filter *.meta files
        # TODO: Use ssh to ask rutorrent for unfinished downloads
        with self.reserve_ftp() as ftp:
            directory_info = DirectoryInfo(ftp, directory, recursive=False)
        return directory_info.all_paths

    def reserve_ftp(self) -> Generator[ftplib.FTP, None, None]:
        return self._ftp_pool.reserve()

    def download_file(self, file: FileInfo, callback: Callable = lambda _: None) -> None:
        output_file = self._get_download_path(file.path)
        with output_file.open('wb') as output:
            with self.reserve_ftp() as ftp:
                def _callback(data):
                    output.write(data)
                    callback(data)
                ftp.retrbinary(f'RETR {file.path}', _callback)

    def _file_type(self, file: Path) -> FileType:
        with self.reserve_ftp() as ftp:
            if 'type=dir' in ftp.sendcmd(f'MLST {file}'):
                return FileType.DIR
            else:
                return FileType.FILE

    def _get_directory_size(self, directory: Path) -> int:
        with self.reserve_ftp() as ftp:
            directory_info = DirectoryInfo(ftp, directory)
        size = sum([f.size for f in directory_info.files])
        return size

    def _get_download_path(self, path: Path) -> Path:
        return fs.replace_path_prefix(path, self.base_dir, self.download_dir)

    def _get_file_size(self, file: Path) -> int:
        with self.reserve_ftp() as ftp:
            ftp.voidcmd('TYPE I')
            size = ftp.size(str(file))
        return size

    def _make_download_dirs(self, path: Path, directory_info: DirectoryInfo) -> None:
        fs.make_dir(self._get_download_path(path))
        for directory in directory_info.directories:
            fs.make_dir(self._get_download_path(directory))



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .single_argument import SingleArgument


class CountArgument(SingleArgument):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, metavar='<n>', default=None, type=int, **kwargs)



"""git constants"""

from typing import Dict

GitConfig = Dict[str, str]

HEAD: str = 'HEAD'
ORIGIN: str = 'origin'
FETCH_URL: str = 'fetch_url'
PUSH_URL: str = 'push_url'
GITMODULES: str = '.gitmodules'



"""progress utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from threading import Lock
from typing import Any, Dict, Optional

from rich.console import Console
from rich.progress import TaskID, BarColumn, ProgressColumn, Task
from rich.progress import Progress as RichProgress

from .format import Format
from .console import CONSOLE


class DownloadProgressColumn(ProgressColumn):

    def __init__(self) -> None:
        super().__init__()
        self._width: int = 0

    def render(self, task: Task) -> str:
        if 'units' in task.fields and task.fields['units'] == 'bytes':
            downloaded = Format.gnu_size(int(task.completed))
            total = Format.gnu_size(int(task.total))
            output = f'{downloaded}/{total}'
        else:
            output = f'{task.completed}/{task.total}'

        width = max(self._width, len(output))
        if width > self._width:
            self._width = width
        return output.rjust(self._width)


class Progress:

    def __init__(self, console: Optional[Console] = None, should_clear_lines: bool = True):
        self._delete_line_count: int = 0
        self._lock: Lock = Lock()
        self._task_ids: Dict[str: TaskID] = {}
        self._subtask_ids: Dict[str: TaskID] = {}
        self._should_clear_lines: bool = should_clear_lines

        self._progress_bar_format = [
            "[progress.description]{task.description}",
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.0f}%",
            DownloadProgressColumn(),
            "{task.fields[units]}"
        ]
        self._progress: RichProgress = RichProgress(*self._progress_bar_format, console=console)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def add_task(self, identifier: Any, total: int, units: str, start: bool = True) -> None:
        task_id = self._progress.add_task(str(identifier), total=total, units=units, start=start)
        self._add_task_id(identifier, task_id)

    def add_subtask(self, identifier: Any, total: int, units: str, start: bool = True) -> None:
        task_id = self._progress.add_task(str(identifier), total=total, units=units, start=start)
        self._add_subtask_id(identifier, task_id)
        count = len(self._progress.tasks)
        if count <= 1:
            return
        with self._lock:
            self._delete_line_count = max(self._delete_line_count, count - 1)

    def clear_lines(self) -> None:
        if self._should_clear_lines:
            CONSOLE.delete_lines(count=self._delete_line_count, force=True)

    def complete_task(self, identifier: Any) -> None:
        task_id = self._get_task_id(identifier)
        task = self._progress.tasks[task_id]
        self._progress.update(task_id, completed=task.total)

    def complete_subtask(self, identifier: Any) -> None:
        task_id = self._get_subtask_id(identifier)
        # task = self._progress.tasks[task_id]
        # self._progress.update(task_id, completed=task.total, visible=False)
        self._progress.remove_task(task_id)

    def start(self) -> None:
        self._progress.start()

    def start_task(self, identifier: Any) -> None:
        task_id = self._get_task_id(identifier)
        self._progress.start_task(task_id)

    def start_subtask(self, identifier: Any) -> None:
        task_id = self._get_subtask_id(identifier)
        self._progress.start_task(task_id)

    def stop(self, clear_lines: bool = True) -> None:
        self._progress.stop()
        if clear_lines:
            self.clear_lines()

    def update_task(self, identifier: Any, advance: int) -> None:
        task_id = self._get_task_id(identifier)
        self._progress.update(task_id, advance=advance)

    def update_subtask(self, identifier: Any, advance: int) -> None:
        task_id = self._get_subtask_id(identifier)
        self._progress.update(task_id, advance=advance)

    def _get_task_id(self, identifier: Any) -> TaskID:
        return self._task_ids[str(identifier)]

    def _get_subtask_id(self, identifier: Any) -> TaskID:
        return self._subtask_ids[str(identifier)]

    def _add_task_id(self, identifier: Any, task_id: TaskID) -> None:
        with self._lock:
            self._task_ids[str(identifier)] = task_id

    def _add_subtask_id(self, identifier: Any, task_id: TaskID) -> None:
        with self._lock:
            self._subtask_ids[str(identifier)] = task_id



"""console utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import sys
from functools import wraps
from io import StringIO
from typing import Any, List, Optional

from rich.console import Console as RichConsole


def disable_output(func):
    """Disable Console output for wrapped function"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper"""

        CONSOLE.print_output = False
        try:
            retval = func(*args, **kwargs)
            CONSOLE.print_output = True
            return retval
        except:  # noqa
            CONSOLE.print_output = True
            raise

    return wrapper


class Console:
    """Console class"""

    def __init__(self):
        self._queue: List[str] = []
        self.print_output: bool = True
        self._stdout: RichConsole = RichConsole(force_terminal=True,
                                                # color_system='256',
                                                width=160)
        self._stderr: RichConsole = RichConsole(file=sys.stderr,
                                                force_terminal=True,
                                                # color_system='256',
                                                width=160)
        self._stringio: RichConsole = RichConsole(file=StringIO(),
                                                  width=160)

    @property
    def stdout_console(self) -> RichConsole:
        return self._stdout

    def stderr(self, output: Any = '', force: bool = False) -> None:
        if output is None:
            return
        if self.print_output or force:
            self._stderr.log(output)

    def stdout(self, output: Any = '', force: bool = False) -> None:
        if output is None:
            return
        if self.print_output or force:
            self._stdout.print(output)

    def print_exception(self, force: bool = False) -> None:
        if self.print_output or force:
            self._stderr.print_exception()

    def delete_lines(self, count: int = 1, force: bool = False) -> None:
        if self.print_output or force:
            for _ in range(count):
                print('\033[A                             \033[A')

    def enqueue_stdout(self, output: Any = '', newline: bool = False) -> None:
        if output is None:
            return
        if isinstance(output, list):
            self._queue += output
        elif isinstance(output, str):
            self._queue.append(output)

        if newline:
            self._queue.append('')

    def flush_stdout(self, force: bool = False) -> None:
        if self.print_output or force:
            self._stdout.print('\n'.join(self._queue))
        self._queue = []

    @property
    def width(self) -> int:
        width = self._stdout.width
        return width

    @staticmethod
    def _debug_console(console: RichConsole) -> None:
        print(console.encoding)
        print(console.color_system)
        print(console.width)
        print(console.is_terminal)
        print(console.is_dumb_terminal)

    def _pretty_log_message(self, output: Any) -> Optional[str]:
        if output is None:
            return None
        self._stringio.print(output)
        output = self._stringio.file.getvalue()
        return output

    @property
    def _pretty_traceback(self) -> str:
        self._stringio.print_exception()
        output = self._stringio.file.getvalue()
        return output


CONSOLE: Console = Console()



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from typing import Any, Dict, Tuple

from ..format import Format


class Argument:

    def __init__(self, *args, **kwargs):
        self.args: Tuple[Any] = args

        # update metavar to default if missing or remove if None
        metavar = 'metavar'
        if metavar not in kwargs:
            name = self._get_name(args)
            kwargs[metavar] = f'<{name}>'
        elif metavar in kwargs and kwargs[metavar] is None:
            del kwargs[metavar]

        self.options: Dict[str, Any] = kwargs

    @staticmethod
    def _get_name(args: Tuple[Any]) -> str:
        names = [a for a in args if a.startswith('--')]
        if names:
            return Format.remove_prefix(names[0], '--')

        names = [a for a in args if a.startswith('-')]
        if names:
            return Format.remove_prefix(names[0], '-')

        names = [a for a in args if not a.startswith('-')]
        if names:
            return names[0]

        raise Exception('Failed to infer argument name')



"""error

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""


class ExistingFileError(Exception):
    pass


class ExistingSymlinkError(Exception):
    pass


class MissingFileError(Exception):
    pass


class UnknownArgumentError(Exception):
    pass


class GitError(Exception):
    pass



"""general utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

# from abc import abstractmethod
# from typing import Protocol
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, MutableSequence


# class Comparable(Protocol):
#     """Protocol for annotating comparable types."""
#
#     @abstractmethod
#     def __lt__(self: 'CT', other: 'CT') -> bool:
#         pass


T = TypeVar('T')
# CT = TypeVar('CT', bound=Comparable)


def is_even(number: int) -> bool:
    return (number % 2) == 0


def sorted_tuple(items: MutableSequence[T], unique: bool = False, reverse: bool = False,
                 key: Optional[Callable] = None) -> Tuple[T, ...]:
    if unique:
        items = set(items)
    return tuple(sorted(items, reverse=reverse, key=key))


def values_sorted_by_key(dictionary: Dict[Any, T]) -> List[T]:
    return [dictionary[key] for key in sorted(dictionary.keys())]



"""Clowder command line utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

import jsonschema
import yaml as pyyaml

import pygoodle.filesystem as fs
from pygoodle.console import CONSOLE
from .error import ExistingFileError
from .format import Format

YAML: str = '.yaml'
YML: str = '.yml'


class InvalidYamlError(Exception):
    pass


class MissingYamlError(Exception):
    pass


class Yaml:

    def __init__(self, path: Path, schema: Optional[str] = None):
        if path.suffix != YAML and path.suffix != YML:
            path = path / YML
        self.path: Path = path
        self.schema: Optional[str] = schema

    @property
    def exists(self) -> bool:
        return self.path.is_file()

    @property
    def path_with_alternate_extension(self) -> Path:
        if self.has_yaml_extension:
            return self.path_with_yml_extension
        else:
            return self.path_with_yaml_extension

    @property
    def alternate_extension_exists(self) -> bool:
        return self.path_with_alternate_extension.is_file()

    @property
    def has_yaml_extension(self) -> bool:
        return self.path.suffix == YAML

    @property
    def has_yml_extension(self) -> bool:
        return self.path.suffix == YML

    @property
    def path_with_yaml_extension(self) -> Path:
        if self.has_yaml_extension:
            return self.path
        return self.path.parent / f'{self.path.stem}{YAML}'

    @property
    def path_with_yml_extension(self) -> Path:
        if self.has_yml_extension:
            return self.path
        return self.path.parent / f'{self.path.stem}{YML}'

    def delete(self, alternate_extension: bool = False):
        if not alternate_extension:
            fs.remove(self.path)
            return

        if self.has_yaml_extension:
            fs.remove(self.path_with_yml_extension)
        else:
            fs.remove(self.path_with_yaml_extension)

    def load(self, relative_to: Optional[Path] = None) -> dict:
        """Load clowder config from yaml file

        :param Optional[Path] relative_to: Directory yaml file is relative to
        :return: YAML python object
        :raise InvalidYamlError:
        """

        try:
            with self.path.open() as raw_file:
                parsed_yaml = pyyaml.safe_load(raw_file)
                if parsed_yaml is None:
                    raise InvalidYamlError(f"No entries in {Format.path(self.path, relative_to=relative_to)}")
                return parsed_yaml
        except pyyaml.YAMLError:
            # LOG.error(f"Failed to open file '{yaml_file}'")
            raise

    def save(self, contents: dict) -> None:
        """Save yaml file to disk

        :param dict contents: Parsed YAML python object
        :raise ExistingFileError:
        """

        if self.path.is_file():
            raise ExistingFileError(f"File already exists: {Format.path(self.path)}")

        CONSOLE.stdout(f" - Save yaml to file at {Format.path(self.path)}")
        try:
            with self.path.open(mode="w") as raw_file:
                pyyaml.safe_dump(contents, raw_file, default_flow_style=False, indent=2, sort_keys=False)
        except pyyaml.YAMLError:
            # LOG.error(f"Failed to save file {Format.path(yaml_file)}")
            raise

    def validate(self, schema: Optional[str] = None, relative_to: Optional[Path] = None) -> dict:
        """Validate yaml file

        :param str schema: json schema
        :param Optional[Path] relative_to: Path to load relative to
        :return: Parsed YAML python object
        """

        schema = self.schema if schema is None else schema
        try:
            schema = pyyaml.safe_load(schema)
            parsed = self.load(relative_to=relative_to)
            jsonschema.validate(parsed, schema)
            return parsed
        except jsonschema.exceptions.ValidationError:
            # LOG.error(f'Yaml json schema validation failed {Format.invalid_yaml(file_path.name)}\n')
            raise

    @staticmethod
    def get_string(dictionary: dict) -> str:
        """Return yaml string from python data structures

        :param dict dictionary: YAML python object
        :return: YAML as a string
        """

        try:
            return pyyaml.safe_dump(dictionary, default_flow_style=False, indent=2, sort_keys=False).strip()
        except pyyaml.YAMLError:
            # LOG.error(f"Failed to dump yaml file contents",)
            raise

    def update_extension(self) -> Optional[Path]:
        if self.path_with_yml_extension.exists():
            self.path = self.path_with_yml_extension
        elif self.path_with_yaml_extension.exists():
            self.path = self.path_with_yaml_extension
        return self.path



"""Subprocess command utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import os
import subprocess
from pathlib import Path
from subprocess import CompletedProcess, PIPE, STDOUT
from typing import List, Optional, Union

from .console import CONSOLE
from .format import Format


def get_stdout(command: str, cwd: Path = Path.cwd()) -> Optional[str]:
    if not cwd.is_dir():
        return None
    result = run(command, cwd=cwd, print_output=False, check=False)
    if result.returncode != 0:
        return None
    output: str = result.stdout
    output = output.strip()
    if not output:
        return None
    return output


def run_silent(command: str, cwd: Path = Path.cwd()) -> CompletedProcess:
    return run(command, cwd=cwd, check=False, print_output=False)


def run(command: Union[str, List[str]], cwd: Path = Path.cwd(), check: bool = True,
        env: Optional[dict] = None, stdout=PIPE, stderr=STDOUT,
        print_output: Optional[bool] = None, print_command: bool = False,
        login: bool = False, interactive: bool = False, executable: Optional[str] = None) -> CompletedProcess:

    if print_output is None:
        print_output = CONSOLE.print_output

    if print_output:
        stdout = None
        stderr = None

    if print_command:
        output = Format.default(f"> {command}")
        output = Format.bold(output)
        CONSOLE.stdout(output)

    # if isinstance(command, list):
    #     cmd = ' '.join(command)
    # else:
    #     cmd = command

    if isinstance(command, str):
        command = [command]
    if login:
        command = ['-l'] + command
    if interactive:
        command = ['-i'] + command

    cmd_env = os.environ.copy()
    if env is not None:
        cmd_env.update(env)

    if executable is not None:
        executable = executable
        cmd_env['SHELL'] = executable

    # TODO: Replace universal_newlines with text when Python 3.6 support is dropped
    completed_process = subprocess.run(
        command,
        cwd=cwd,
        env=cmd_env,
        shell=True,
        stdout=stdout,
        stderr=stderr,
        universal_newlines=True,
        check=check,
        executable=executable
    )

    return completed_process



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import argparse
from typing import List, Optional, Type

import pygoodle.reflection as reflect
from .argument import Argument
from .argument_group import ArgumentGroup
from .mutually_exclusive_argument_group import MutuallyExclusiveArgumentGroup


class Subcommand:

    class Meta:
        name: str = 'subcommand'
        help: str = f'{name} help'
        args: List[Argument] = []
        mutually_exclusive_args: List[MutuallyExclusiveArgumentGroup] = []
        argument_groups: List[ArgumentGroup] = {}
        subcommands: List[Type['Subcommand']] = []

    def __init__(self):
        self._parser: Optional[argparse.ArgumentParser] = None
        self.name: str = 'subcommand'
        self.help: str = f'{self.name} help'
        self.args: List[Argument] = []
        self.mutually_exclusive_args: List[MutuallyExclusiveArgumentGroup] = []
        self.argument_groups: List[ArgumentGroup] = []
        self.subcommands: List[Type['Subcommand']] = []
        self._update_meta()

    @staticmethod
    def run(args) -> None:
        raise NotImplementedError

    def print_help(self) -> None:
        self._parser.print_help()

    def add_parser(self, parser) -> None:
        self._parser = parser

    def _update_attr(self, name: str, meta: Meta) -> None:
        reflect.update_attr(self, name, meta)

    def _update_meta(self) -> None:
        classes = reflect.method_resolution_order(self, reverse=True)
        for cls in classes:
            meta = reflect.class_member(cls, 'Meta')
            if meta is not None:
                self._update_attr('name', meta)
                self._update_attr('help', meta)
                self._update_attr('args', meta)
                self._update_attr('mutually_exclusive_args', meta)
                self._update_attr('argument_groups', meta)
                self._update_attr('subcommands', meta)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from enum import auto, unique

from pygoodle.enum import AutoLowerName, UnknownEnumCaseError


@unique
class Protocol(AutoLowerName):
    SSH = auto()
    HTTPS = auto()

    def format_url(self, url: str, name: str) -> str:
        """Return formatted git url

        :param str url: Repo url
        :param str name: Repo name
        :return: Full git repo url for specified protocol
        :raise UnknownTypeError:
        """

        if self is Protocol.SSH:
            return f"git@{url}:{name}.git"
        elif self is Protocol.HTTPS:
            return f"https://{url}/{name}.git"
        else:
            raise UnknownEnumCaseError('Invalid git protocol')



"""reflection utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

# import inspect
from typing import Any, List, Optional, Type


def all_subclasses(cls) -> List[Type]:
    subclasses = []
    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(all_subclasses(subclass))
    return subclasses


def class_member(obj: Any, name: str) -> Optional[Any]:
    class_dict = obj.__dict__
    if name in class_dict:
        return class_dict[name]
    return None


def update_attr(obj: Any, name: str, value: Any, ignore_missing: bool = True):
    has_attr = hasattr(value, name)
    if not ignore_missing and not has_attr:
        raise Exception(f'Missing attribute {name} on {value}')
    if has_attr:
        meta_value = getattr(value, name)
        setattr(obj, name, meta_value)


def method_resolution_order(obj, reverse: bool = False) -> List[Type]:
    # classes = inspect.getmro(type(obj))
    classes = type.mro(type(obj))
    if not reverse:
        return classes
    classes.reverse()
    return classes



"""Misc git utils"""

from typing import Dict, List, Optional, Tuple

from pygoodle.format import Format

from .constants import FETCH_URL, PUSH_URL

# TODO: Update to use ConfigParser


class ProcessOutput:

    @classmethod
    def tag_shas(cls, output: str) -> Dict[str, str]:
        # TODO: Add expected output example

        return ProcessOutput.shas(output, 'refs/tags/')

    @classmethod
    def branch_shas(cls, output: str) -> Dict[str, str]:
        # Expected output format:
        # 6def4cee3c6abe73ab2889d155421a90722282ef	refs/heads/gh-pages
        # bc787b2888acf4b7bd8351b9a72982c71c143362	refs/heads/master
        # 1ca96862f7814d9ec8b28fff17e913e11add342f	refs/heads/presentation
        # d3db093e48bd685b7eaf8da9cedcef3c9ef48b29	refs/heads/wip/force-sync

        return ProcessOutput.shas(output, 'refs/heads/')

    @classmethod
    def shas(cls, output: str, prefix: str) -> Dict[str, str]:
        # TODO: Add expected output example

        lines = output.strip().splitlines()
        items = {}
        for line in lines:
            components = line.split()
            sha = components[0].strip()
            name = Format.remove_prefix(components[1], prefix)
            items[name] = sha
        return items

    @classmethod
    def local_branches(cls, output: str) -> List[str]:
        # TODO: Add expected output example

        lines = output.splitlines()
        return [line.split()[1].strip() if line.startswith('*') else line.strip() for line in lines]

    @classmethod
    def tracking_branches(cls, output: str) -> Tuple[str, Optional[str]]:
        # Expected output format:
        # > git rev-parse --symbolic-full-name git-old@{upstream}
        # refs/heads/git
        # > git rev-parse --symbolic-full-name git-old@{upstream}
        # refs/remotes/origin/git

        if output.startswith('refs/heads'):
            remote = None
            branch = Format.remove_prefix('refs/heads', output)
        elif output.startswith('refs/remotes'):
            components = Format.remove_prefix('refs/remotes', output).split('/')
            remote = components[0]
            branch = components[1]
        else:
            raise Exception('Failed to parse tracking branch output')
        return branch, remote

    @classmethod
    def remote_branches(cls, output: str, remote: str) -> Tuple[List[str], Optional[str]]:
        # TODO: Add expected output example

        lines = output.strip().splitlines()
        branches = []
        default_branch = None
        for line in lines:
            components = line.split()
            if components[0] == 'warning:':
                continue
            if len(components) == 1:
                name = Format.remove_prefix(components[0].strip(), f'{remote}/')
                branches.append(name)
            elif len(components) == 3 and components[1] == '->':
                name = Format.remove_prefix(components[2].strip(), f'{remote}/')
                default_branch = name
            else:
                raise Exception('Wrong number of components for remote branch')
        return branches, default_branch

    @classmethod
    def remotes(cls, output: str) -> Dict[str, Dict[str, str]]:
        # Expected output format:
        # origin	git@github.com:JrGoodle/pygoodle.git (fetch)
        # origin	git@github.com:JrGoodle/pygoodle.git (push)

        lines = output.splitlines()
        line = [line.split() for line in lines]
        remotes = {}
        for components in line:
            name = components[0].strip()
            url = components[1].strip()
            kind = components[2].strip()
            if name not in remotes.keys():
                remotes[name] = {}
            if kind == '(fetch)':
                remotes[name][FETCH_URL] = url
            elif kind == '(push)':
                remotes[name][PUSH_URL] = url
            else:
                raise Exception('Unknown')
        return remotes

    @classmethod
    def submodules(cls, output: List[str]) -> Dict[str, Dict[str, str]]:
        # Expected output format for .gitmodules:
        # submodule.path/to/Optional.path path/to/Optional
        # submodule.path/to/Optional.url https://github.com/akrzemi1/Optional.git
        # submodule.path/to/Optional.branch master
        #
        # Expected output format for .git/config:
        # submodule.path/to/Optional.active true
        # submodule.path/to/Optional.url https://github.com/akrzemi1/Optional.git

        submodules = {}
        for submodule_info in output:
            info_components = submodule_info.split()
            name = info_components[0]
            value = info_components[1]
            name_components = name.split('.')
            if len(name_components) != 3 or name_components[0] != 'submodule':
                continue
            submodule_path = name_components[1]
            key = name_components[2]
            if submodule_path not in submodules.keys():
                submodules[submodule_path] = {}
            submodules[submodule_path][key] = value
        return submodules

    @classmethod
    def diff_index(cls, output: str) -> Dict[str, List[Dict[str, str]]]:
        # Expected output format
        # :000000 100644 0000000000000000000000000000000000000000 e69de29bb2d1d6434b8b29ae775ad8c2e48c5391 A	something/something.txt
        # :100644 100644 375928ea14a5a1a7e14f14a36cb0a31c417d1ca1 0000000000000000000000000000000000000000 M	README.md
        # :100644 000000 375928ea14a5a1a7e14f14a36cb0a31c417d1ca1 0000000000000000000000000000000000000000 D	README.md

        changes = {
            'added': [],
            'modified': [],
            'deleted': []
        }

        if output is None:
            return changes

        def diff_type(value: str) -> str:
            if value == 'A':
                return 'added'
            elif value == 'M':
                return 'modified'
            elif value == 'D':
                return 'deleted'
            else:
                raise Exception('Unknown diff type')

        lines = output.strip().splitlines()
        for line in lines:
            old_permissions = Format.remove_prefix(line[0].strip(), ':')
            new_permissions = line[1].strip()
            old_sha = line[2].strip()
            new_sha = line[3].strip()
            change_type = line[4].strip()
            file_path = line[5].strip()
            change = {
                'old_permissions': old_permissions,
                'new_permissions': new_permissions,
                'old_sha': old_sha,
                'new_sha': new_sha,
                'path': file_path
            }
            changes[diff_type(change_type)].append(change)
        return changes



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pygoodle.console import CONSOLE
from pygoodle.format import Format
# from pygoodle.git.decorators import error_msg
from pygoodle.git.offline import GitOffline

from .tag import Tag


class LocalTag(Tag):
    """Class encapsulating git tag

    :ivar str name: Branch
    :ivar str formatted_ref: Formatted ref
    """

    @property
    def is_checked_out(self) -> bool:
        current_sha = GitOffline.current_head_commit_sha(self.path)
        tag_sha = GitOffline.get_tag_commit_sha(self.path, self.name)
        return current_sha == tag_sha

    @property
    def sha(self) -> str:
        """Commit sha"""
        return GitOffline.get_tag_commit_sha(self.path, self.name)

    def create(self) -> None:
        if self.exists:
            CONSOLE.stdout(f' - Local tag {Format.Git.ref(self.name)} already exists')
            return
        raise NotImplementedError

    # @error_msg('Failed to delete local tag')
    def delete(self) -> None:
        if not self.exists:
            CONSOLE.stdout(f" - Local tag {Format.Git.ref(self.short_ref)} doesn't exist")
            return
        CONSOLE.stdout(f' - Delete local tag {Format.Git.ref(self.short_ref)}')
        GitOffline.delete_local_tag(self.path, name=self.name)

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_local_tag(self.path, self.name)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path

from pygoodle.console import CONSOLE
from pygoodle.git.model.ref import Ref
from pygoodle.git.offline import GitOffline


class Tag(Ref):
    """Class encapsulating git tag

    :ivar Path path: Path to git repo
    :ivar str name: Branch
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path, name: str):
        """GitRepo __init__

        :param Path path: Path to git repo
        :param str name: Tag name
        """

        super().__init__(path)
        self.name: str = name
        self.check_ref_format(self.formatted_ref)

    def __eq__(self, other) -> bool:
        if isinstance(other, Tag):
            return self.name == other.name and self.path == other.path
        return False

    def __lt__(self, other: 'Tag') -> bool:
        return self.name < other.name

    @property
    def is_checked_out(self) -> bool:
        raise NotImplementedError

    @property
    def is_tag(self) -> bool:
        return True

    @property
    def sha(self) -> str:
        """Commit sha"""
        raise NotImplementedError

    @property
    def short_ref(self) -> str:
        """Short git ref"""

        return self.truncate_ref(self.name)

    @property
    def formatted_ref(self) -> str:
        """Formatted git ref"""

        return self.format_git_tag(self.name)

    def checkout(self, check: bool = True, track: bool = False) -> None:
        current_commit = GitOffline.current_head_commit_sha(self.path)
        if current_commit == self.sha:
            CONSOLE.stdout(' - On correct commit for tag')
            return
        CONSOLE.stdout(f' - Checkout tag {self.name}')
        super().checkout(check=check, track=track)



"""random utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import random
import string


def random_alphanumeric_string(length: int) -> str:
    letters_and_digits = string.ascii_letters + string.digits
    result = ''.join((random.choice(letters_and_digits) for _ in range(length)))
    return result



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.git.constants import ORIGIN
# from pygoodle.git.decorators import not_detached
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline
from pygoodle.format import Format
# from pygoodle.git.decorators import error_msg

from .branch import Branch


class TrackingBranch(Branch):
    """Class encapsulating git branch

    :ivar Path path: Path to git repo
    :ivar str name: Local branch name
    :ivar RemoteBranch upstream_branch: Upstream branch name
    :ivar Optional[RemoteBranch] push_branch: Push branch name
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path, local_branch: str,
                 upstream_branch: Optional[str] = None, upstream_remote: Optional[str] = None,
                 push_branch: Optional[str] = None, push_remote: Optional[str] = None):
        super().__init__(path, local_branch)
        from .local_branch import LocalBranch
        from .remote_branch import RemoteBranch
        upstream_branch = local_branch if upstream_branch is None else upstream_branch
        upstream_remote = ORIGIN if upstream_remote is None else upstream_remote
        push_branch = upstream_branch if push_branch is None else push_branch
        push_remote = upstream_remote if push_remote is None else push_remote
        self.local_branch: LocalBranch = LocalBranch(self.path, self.name)
        self.upstream_branch: RemoteBranch = RemoteBranch(self.path, upstream_branch, upstream_remote)
        self.push_branch: RemoteBranch = RemoteBranch(self.path, push_branch, push_remote)

    def __eq__(self, other) -> bool:
        if isinstance(other, TrackingBranch):
            return super().__eq__(other) and self.upstream_branch == other.upstream_branch
        return False

    def __lt__(self, other: 'TrackingBranch') -> bool:
        return f'{self.name}/{self.upstream_branch.remote.name}/{self.upstream_branch.name}' \
               < f'{other.name}/{other.upstream_branch.remote.name}/{other.upstream_branch.name}'

    @property
    def sha(self) -> Optional[str]:
        """Commit sha"""
        raise self.local_branch.sha

    def delete(self, force: bool = False) -> None:
        self.local_branch.delete(force=force)
        self.upstream_branch.delete()

    @property
    def is_branch(self) -> bool:
        return True

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_tracking_branch(self.path, self.name)

    # @error_msg('Failed to set tracking branch')
    def set_upstream(self) -> None:
        CONSOLE.stdout(f' - Set tracking branch {Format.Git.ref(self.name)} -> '
                       f'{Format.Git.remote(self.upstream_branch.remote.name)} '
                       f'{Format.Git.ref(self.upstream_branch.name)}')
        GitOffline.set_upstream_branch(self.path,
                                       local_branch=self.name,
                                       upstream_branch=self.upstream_branch.name,
                                       remote=self.upstream_branch.remote.name)

    # @error_msg('Failed to create tracking branch')
    def create(self) -> None:
        if GitOffline.has_tracking_branch(self.path, self.local_branch.name):
            CONSOLE.stdout(' - Tracking branch already exists')
            return
        # TODO: Add Format util to format tracking branch output: local_branch -> remote remote_branch
        CONSOLE.stdout(f' - Create tracking branch {Format.Git.ref(self.name)}')
        GitOnline.fetch(self.path, prune=True)
        # local and remote branches exist
        if self.local_branch.exists and self.upstream_branch.exists:
            self.set_upstream()
            return

        # only local branch exists
        if self.local_branch.exists:
            # GitOnline.push(self.path,
            #                local_branch=self.name,
            #                remote_branch=self.upstream_branch.name,
            #                remote=self.upstream_branch.remote.name,
            #                set_upstream=True)
            self.upstream_branch.create(branch=self.local_branch.name)
            self.set_upstream()
            GitOnline.fetch(self.path, prune=True)
            return

        # only remote branch exists
        if self.upstream_branch.exists:
            self.local_branch.create(branch=self.upstream_branch.name, remote=self.upstream_branch.remote.name)
            # GitOffline.create_local_branch(self.path, self.name,
            #                                branch=self.upstream_branch.name,
            #                                remote=self.upstream_branch.remote)
            return

        # local and remote branches DO NOT exist
        self.upstream_branch.create()
        GitOnline.fetch(self.path, prune=True)
        self.local_branch.create(branch=self.upstream_branch.name, remote=self.upstream_branch.remote.name)

    # @error_msg('Failed to pull')
    # @not_detached
    def pull(self, rebase: bool = False, prune: bool = False, tags: bool = False,
             jobs: Optional[int] = None, no_edit: bool = False, autostash: bool = False,
             depth: Optional[int] = None) -> None:
        # TODO: Check if detached
        self.upstream_branch.pull(rebase=rebase, prune=prune, tags=tags, jobs=jobs,
                                  no_edit=no_edit, autostash=autostash, depth=depth)

    def _set_tracking_branch_commit(self, branch: str, remote: str, depth: int) -> None:
        """Set tracking relationship between local and remote branch if on same commit

        :param str branch: Branch name
        :param str remote: Remote name
        :param int depth: Git clone depth. 0 indicates full clone, otherwise must be a positive integer
        :raise ClowderGitError:
        """

        # FIXME: Try to set this in any scenario where it makes sense
        # origin = self._remote(remote)
        # self.fetch(remote, depth=depth, ref=GitRef(branch=branch))
        #
        # if not self.has_local_branch(branch):
        #     raise ClowderGitError(f'No local branch {fmt.ref(branch)}')
        #
        # if not self.has_remote_branch(branch, remote):
        #     raise ClowderGitError(f'No remote branch {fmt.ref(branch)}')
        #
        # local_branch = self.repo.heads[branch]
        # remote_branch = origin.refs[branch]
        # if local_branch.commit != remote_branch.commit:
        #     raise ClowderGitError(f' - Existing remote branch {fmt.ref(branch)} on different commit')
        #
        # self._set_tracking_branch(remote, branch)



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .argument import Argument


class BoolArgument(Argument):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, action='store_true', metavar=None, **kwargs)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.format import Format
from pygoodle.git.model.ref import Ref
from pygoodle.git.offline import GitOffline


class Branch(Ref):
    """Class encapsulating git branch

    :ivar Path path: Path to git repo
    :ivar str name: Branch name
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path, name: str):
        """Branch __init__

        :param Path path: Path to git repo
        :param str name: Branch name
        """

        super().__init__(path)
        self.name: str = name
        self.check_ref_format(self.formatted_ref)

    def __eq__(self, other) -> bool:
        if isinstance(other, Branch):
            return super().__eq__(other) and self.name == other.name
        return False

    @property
    def is_checked_out(self) -> bool:
        current_branch = GitOffline.current_branch(self.path)
        return current_branch == self.name

    @property
    def is_branch(self) -> bool:
        return True

    @property
    def is_tracking_branch(self) -> bool:
        return GitOffline.has_tracking_branch(self.path, self.name)

    def delete(self) -> None:
        raise NotImplementedError

    @property
    def sha(self) -> Optional[str]:
        """Commit sha"""
        raise NotImplementedError

    @property
    def short_ref(self) -> str:
        """Short git ref"""

        return self.truncate_ref(self.name)

    @property
    def formatted_ref(self) -> str:
        """Formatted git ref"""

        return self.format_git_branch(self.name)

    def checkout(self, check: bool = True, track: bool = False) -> None:
        current_branch = GitOffline.current_branch(self.path)
        if current_branch == self.name:
            CONSOLE.stdout(f' - Branch {Format.Git.ref(self.short_ref)} already checked out')
            return
        CONSOLE.stdout(f' - Checkout branch {Format.Git.ref(self.short_ref)}')
        super().checkout(check=check, track=track)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.format import Format
from pygoodle.git.constants import ORIGIN
# from pygoodle.git.decorators import error_msg
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline
from pygoodle.git.model.remote import Remote

from .tag import Tag


class RemoteTag(Tag):
    """Class encapsulating git tag

    :ivar Path path: Path to git repo
    :ivar str name: Tag name
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path, name: str, remote: Optional[str] = None):
        """GitRepo __init__

        :param str name: Tag name
        :param Remote remote: Remote
        """

        remote = ORIGIN if remote is None else remote
        super().__init__(path, name)
        self.remote: Remote = Remote(self.path, remote)

    def __eq__(self, other) -> bool:
        if isinstance(other, RemoteTag):
            return super().__eq__(other) and self.remote.name == other.remote.name
        return False

    def __lt__(self, other: 'RemoteTag') -> bool:
        return f'{self.remote.name}/{self.name}' < f'{other.remote.name}/{other.name}'

    @property
    def is_checked_out(self) -> bool:
        current_sha = GitOffline.current_head_commit_sha(self.path)
        tag_sha = GitOnline.get_remote_tag_sha(self.path, self.name)
        return current_sha == tag_sha

    @property
    def sha(self) -> str:
        """Commit sha"""
        return GitOnline.get_remote_tag_sha(self.path, self.name, self.remote.name)

    def create(self) -> None:
        if self.exists:
            CONSOLE.stdout(f' - Remote tag {Format.Git.ref(self.name)} already exists')
            return
        raise NotImplementedError

    # @error_msg('Failed to delete remote tag')
    def delete(self) -> None:
        if not self.exists:
            CONSOLE.stdout(f" - Remote tag {Format.Git.ref(self.short_ref)} doesn't exist")
            return
        CONSOLE.stdout(f' - Delete remote tag {Format.Git.ref(self.short_ref)}')
        GitOnline.delete_remote_tag(self.path, tag=self.name, remote=self.remote.name, force=True)

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_remote_tag(self.path, self.name, self.remote.name)

    def exists_online(self, url: str) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_remote_tag(self.path, tag=self.name, remote=self.remote.name, url=url)



"""clowder diff

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import List, Tuple

from pygoodle.git.model.change import Change, ChangeType
from pygoodle.util import sorted_tuple


class Diff:
    """Class encapsulating git diff

    :ivar Path path: Path to git repo
    """

    def __init__(self, path: Path, changes: List[Change]):
        """Ref __init__

        :param Path path: Path to git repo
        """

        self.path: Path = path
        self.added: Tuple[Change, ...] = sorted_tuple([d for d in changes if d.change_type == ChangeType.ADDED])
        self.modified: Tuple[Change, ...] = sorted_tuple([d for d in changes if d.change_type == ChangeType.MODIFIED])
        self.deleted: Tuple[Change, ...] = sorted_tuple([d for d in changes if d.change_type == ChangeType.DELETED])

    @property
    def has_changes(self) -> bool:
        changes = [self.added, self.modified, self.deleted]
        return not any([len(c) > 0 for c in changes])

    @property
    def has_added_files(self) -> bool:
        return bool(self.added)

    @property
    def has_modified_files(self) -> bool:
        return bool(self.modified)

    @property
    def has_deleted_files(self) -> bool:
        return bool(self.deleted)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from datetime import datetime
from pathlib import Path
from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.git.log import GIT_LOG
from pygoodle.git.offline import GitOffline


class Ref:
    """Class encapsulating git ref

    :ivar Path path: Path to git repo
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path):
        """Ref __init__

        :param Path path: Path to git repo
        """

        self.path: Path = path

    def __eq__(self, other) -> bool:
        if isinstance(other, Ref):
            return self.path == other.path
        return False

    @property
    def is_branch(self) -> bool:
        return False

    @property
    def is_tag(self) -> bool:
        return False

    @property
    def is_commit(self) -> bool:
        return False

    @property
    def sha(self) -> Optional[str]:
        """Commit sha"""
        raise NotImplementedError

    @property
    def short_ref(self) -> str:
        """Short git ref"""
        raise NotImplementedError

    @property
    def formatted_ref(self) -> str:
        """Formatted git ref"""

        raise NotImplementedError

    @property
    def date(self) -> Optional[datetime]:
        """Formatted git ref"""

        return GitOffline.get_commit_date(self.path, self.sha)

    @staticmethod
    def truncate_ref(ref: str) -> str:
        """Return bare branch, tag, or sha

        :param str ref: Full pathspec or short ref
        :return: Ref with 'refs/heads/' and 'refs/tags/' prefix removed
        """

        git_branch = "refs/heads/"
        git_tag = "refs/tags/"
        if ref.startswith(git_branch):
            length = len(git_branch)
        elif ref.startswith(git_tag):
            length = len(git_tag)
        else:
            length = 0
        return ref[length:]

    @staticmethod
    def check_ref_format(ref: str) -> bool:
        """Check if git ref is correctly formatted

        :param str ref: Git ref
        :return: True, if git ref is a valid format
        """

        return GitOffline.check_ref_format(ref)

    @staticmethod
    def format_git_branch(branch: str) -> str:
        """Returns properly formatted git branch

        :param str branch: Git branch name
        :return: Branch prefixed with 'refs/heads/'
        """

        prefix = "refs/heads/"
        return branch if branch.startswith(prefix) else f"{prefix}{branch}"

    @staticmethod
    def format_git_tag(tag: str) -> str:
        """Returns properly formatted git tag

        :param str tag: Git tag name
        :return: Tag prefixed with 'refs/heads/'
        """

        prefix = "refs/tags/"
        return tag if tag.startswith(prefix) else f"{prefix}{tag}"

    def checkout(self, check: bool = True, track: bool = False) -> None:
        try:
            GitOffline.checkout(self.path, ref=self.short_ref, track=track)
        except Exception:  # noqa
            message = f'Failed to checkout'
            if check:
                GIT_LOG.error(message)
                raise
            CONSOLE.stdout(f' - {message}')



"""task pool

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from threading import Lock
from typing import Any, List, Optional

import trio

from pygoodle.console import disable_output
from pygoodle.util import values_sorted_by_key


class Task:

    def __init__(self, name: str):
        self.name: str = name
        self._pool: Optional[TaskPool] = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._pool = None
        pass

    @property
    def cancelled(self) -> bool:
        return self._pool.cancelled

    def in_pool(self, pool: 'TaskPool') -> 'Task':
        self._pool = pool
        return self

    def before_task(self) -> None:
        pass

    def after_task(self) -> None:
        pass

    def run(self) -> None:
        raise NotImplementedError


class TaskPool:

    def __init__(self, jobs: Optional[int] = None):
        self._jobs: Optional[int] = jobs
        self._lock: Lock = Lock()
        self._results: Optional[List[Any]] = None
        self.cancelled: bool = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def before_tasks(self, tasks: List[Task]) -> None:
        pass

    def before_task(self, task: Task) -> None:
        pass

    def after_task(self, task: Task) -> None:
        pass

    def after_tasks(self, tasks: List[Task]) -> None:
        pass

    @disable_output
    def run(self, tasks: List[Task]) -> List[Any]:
        return trio.run(self._run, tasks)

    async def _run(self, tasks: List[Task]) -> List[Any]:
        try:
            async with trio.open_nursery() as nursery:
                limit = None if self._jobs is None else trio.CapacityLimiter(self._jobs)
                self.before_tasks(tasks)
                self._results = {}
                index = 0
                try:
                    for task in tasks:
                        if limit is not None:
                            await limit.acquire_on_behalf_of(task.name)
                        nursery.start_soon(self._run_task, index, task, limit, nursery)
                        index += 1
                except BaseException:
                    nursery.cancel_scope.cancel()
                    self.cancelled = True
                    raise
            return values_sorted_by_key(self._results)
        except BaseException:
            self.cancelled = True
            raise
        finally:
            self.after_tasks(tasks)

    async def _run_task(self, index: int, task: Task, limit: Optional[trio.CapacityLimiter],
                        nursery: trio.Nursery) -> Any:
        with task.in_pool(self):
            try:
                self.before_task(task)
                task.before_task()
                result = await trio.to_thread.run_sync(task.run)
                with self._lock:
                    self._results[index] = result
            except BaseException:
                self.cancelled = True
                nursery.cancel_scope.cancel()
                raise
            finally:
                task.after_task()
                self.after_task(task)
                if limit is not None:
                    limit.release_on_behalf_of(task.name)



"""command line app

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .argument import Argument


class SingleArgument(Argument):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, nargs=1, **kwargs)



"""pygoodle module __init__

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .app import App
from .argument import Argument
from .argument_group import ArgumentGroup
from .bool_argument import BoolArgument
from .count_argument import CountArgument
from .mutually_exclusive_argument_group import MutuallyExclusiveArgumentGroup
from .single_argument import SingleArgument
from .subcommand import Subcommand



"""progress task pool

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from typing import List, Optional

from pygoodle.console import CONSOLE, Console
from pygoodle.progress import Progress

from .task_pool import Task, TaskPool


class ProgressTask(Task):

    def __init__(self, name: str, total: int = 1, units: str = '', start: bool = True):
        super().__init__(name)
        self.progress: Optional[Progress] = None
        self.total: int = total
        self.units: str = units
        self.start: bool = start

    def before_task(self) -> None:
        super().before_task()
        if self.progress is not None:
            self.progress.add_subtask(self.name, total=self.total, units=self.units, start=self.start)

    def after_task(self) -> None:
        super().after_task()
        if self.progress is not None and not self.cancelled:
            self.progress.complete_subtask(self.name)

    def run(self) -> None:
        raise NotImplementedError


class ProgressTaskPool(TaskPool):

    def __init__(self, title: str, units: str = '', jobs: Optional[int] = None,
                 console: Console = CONSOLE.stdout_console, print_subprogress: bool = True):
        super().__init__(jobs)
        self._title: str = title
        self._units = units
        self.progress: Progress = Progress(console=console)
        self._print_subprogress: bool = print_subprogress

    def __enter__(self):
        self.progress.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.progress.stop()

    def before_tasks(self, tasks: List[ProgressTask]) -> None:
        super().before_tasks(tasks)
        if self._print_subprogress:
            for task in tasks:
                task.progress = self.progress
        self.progress.start()
        self.progress.add_task(self._title, total=len(tasks), units=self._units)

    def after_task(self, task: ProgressTask) -> None:
        super().after_task(task)
        if not self.cancelled:
            self.progress.update_task(self._title, advance=1)

    def after_tasks(self, tasks: List[ProgressTask]) -> None:
        super().after_tasks(tasks)
        if not self.cancelled:
            self.progress.complete_task(self._title)
        self.progress.stop(clear_lines=not self.cancelled)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.format import Format
# from pygoodle.git.decorators import error_msg
from pygoodle.git.model.commit import Commit
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .branch import Branch


class LocalBranch(Branch):
    """Class encapsulating git branch

    :ivar Path path: Path to git repo
    :ivar str name: Branch name
    """

    def __lt__(self, other: 'LocalBranch') -> bool:
        return self.name < other.name

    @property
    def is_tracking_branch(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        branches = GitFactory.get_tracking_branches(self.path)
        return any([branch.name == self.name for branch in branches])

    @property
    def sha(self) -> Optional[str]:
        """Commit sha"""
        return GitOffline.get_branch_sha(self.path, self.name)

    # @error_msg('Failed to create local branch')
    def create(self, branch: Optional[str] = None, remote: Optional[str] = None, track: bool = True) -> None:
        if self.exists:
            CONSOLE.stdout(f' - Local branch {Format.Git.ref(self.short_ref)} already exists')
            return
        CONSOLE.stdout(f' - Create local branch {Format.Git.ref(self.short_ref)}')
        if branch is not None:
            if remote is not None:
                from pygoodle.git.model.branch.remote_branch import RemoteBranch
                remote_branch = RemoteBranch(self.path, branch, remote)
                if not remote_branch.exists:
                    branch = None
                    remote = None
            else:
                local_branch = LocalBranch(self.path, branch)
                if not local_branch.exists:
                    branch = None
        GitOffline.create_local_branch(self.path, self.name, branch=branch, remote=remote, track=track)

    # @error_msg('Failed to delete local branch')
    def delete(self, force: bool = False) -> None:
        if not self.exists:
            CONSOLE.stdout(f" - Local branch {Format.Git.ref(self.short_ref)} doesn't exist")
            return
        CONSOLE.stdout(f' - Delete local branch {Format.Git.ref(self.short_ref)}')
        GitOffline.delete_local_branch(self.path, self.name, force=force)

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_local_branch(self.path, self.name)

    @property
    def commit(self) -> Commit:
        sha = GitOffline.get_branch_sha(self.path, branch=self.name)
        return Commit(self.path, sha)

    # @error_msg('Failed to push local changes')
    # @not_detached
    def push(self, remote: Optional[str] = None, branch: Optional[str] = None,
             force: bool = False, set_upstream: bool = False) -> None:
        # TODO: Check if detached
        CONSOLE.stdout(' - Push local changes')
        GitOnline.push(self.path,
                       local_branch=self.name,
                       remote_branch=branch,
                       remote=remote,
                       force=force,
                       set_upstream=set_upstream)



"""clowder ref enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

from pygoodle.console import CONSOLE
from pygoodle.git.constants import ORIGIN
# from pygoodle.git.decorators import not_detached
from pygoodle.format import Format
# from pygoodle.git.decorators import error_msg
from pygoodle.git.model.commit import Commit
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .branch import Branch


class RemoteBranch(Branch):
    """Class encapsulating git branch

    :ivar Path path: Path to git repo
    :ivar str name: Branch name
    :ivar str formatted_ref: Formatted ref
    """

    def __init__(self, path: Path, name: str, remote: Optional[str] = None, is_default: bool = False):
        """Branch __init__

        :param str name: Branch name
        :param Remote remote: Remote
        :param bool is_default: Is branch default for remote repo
        """

        from pygoodle.git.model.remote import Remote
        super().__init__(path, name)
        remote = ORIGIN if remote is None else remote
        self.remote: Remote = Remote(self.path, remote)
        self.is_default: bool = is_default

    def __eq__(self, other) -> bool:
        if isinstance(other, RemoteBranch):
            return super().__eq__(other) and self.remote.name == other.remote.name
        return False

    def __lt__(self, other: 'RemoteBranch') -> bool:
        return f'{self.remote.name}/{self.name}' < f'{other.remote.name}/{other.name}'

    @property
    def short_ref(self) -> str:
        """Short git ref"""

        return f'{self.remote.name}/{self.truncate_ref(self.name)}'

    # TODO: Fix this
    # @property
    # def formatted_ref(self) -> str:
    #     """Formatted git ref"""
    #
    #     return f'refs/remotes/{self.remote.name}/heads/{self.name}'

    @property
    def is_tracking_branch(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_tracking_branch(self.path, self.name)

    @property
    def sha(self) -> Optional[str]:
        """Commit sha"""
        return GitOffline.get_branch_sha(self.path, self.name, self.remote.name)

    # @error_msg('Failed to delete remote branch')
    def delete(self) -> None:
        from pygoodle.git.model.factory import GitFactory
        if not GitFactory.has_remote_branch_online(self.path, self.name, self.remote.name):
            CONSOLE.stdout(f" - Remote branch {Format.Git.ref(self.name)} doesn't exist")
            return
        CONSOLE.stdout(f' - Delete remote branch {Format.Git.ref(self.name)}')
        GitOnline.delete_remote_branch(self.path, branch=self.name, remote=self.remote.name, force=True)

    @property
    def exists(self) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_remote_branch_offline(self.path, self.name, self.remote.name)

    def exists_online(self, url: Optional[str] = None) -> bool:
        from pygoodle.git.model.factory import GitFactory
        return GitFactory.has_remote_branch_online(self.path, branch=self.name, remote=self.remote.name, url=url)

    # @error_msg('Failed to create remote branch')
    def create(self, branch: Optional[str] = None, remote: Optional[str] = None) -> None:
        from pygoodle.git.model.factory import GitFactory
        if GitFactory.has_remote_branch_online(self.path, self.name, self.remote.name):
            CONSOLE.stdout(f' - Remote branch {Format.Git.ref(self.name)} already exists')
            return
        CONSOLE.stdout(f' - Create remote branch {Format.Git.ref(self.name)}')
        branch = self.name if branch is None else branch
        from pygoodle.git.model.branch.local_branch import LocalBranch
        local_branch = LocalBranch(self.path, branch)
        GitOnline.fetch(self.path, prune=True)
        has_existing_branch = local_branch.exists
        if not has_existing_branch:
            local_branch.create(branch=branch, remote=remote)
        local_branch.push(remote=self.remote.name, branch=self.name)
        if not has_existing_branch:
            GitOffline.delete_local_branch(self.path, branch)
        GitOnline.fetch(self.path, prune=True)

    @property
    def commit(self) -> Commit:
        sha = GitOffline.get_branch_sha(self.path, branch=self.name, remote=self.name)
        return Commit(self.path, sha)

    # @error_msg('Failed to pull')
    # @not_detached
    def pull(self, rebase: bool = False, prune: bool = False, tags: bool = False,
             jobs: Optional[int] = None, no_edit: bool = False, autostash: bool = False,
             depth: Optional[int] = None) -> None:
        # TODO: Check if detached
        message = f' - Pull'
        if rebase:
            message += ' with rebase'
        message += f' from {Format.Git.remote(self.remote.name)} {Format.Git.ref(self.name)}'
        CONSOLE.stdout(message)
        GitOnline.pull(self.path, remote=self.remote.name, branch=self.name, rebase=rebase, prune=prune,
                       tags=tags, jobs=jobs, no_edit=no_edit, autostash=autostash, depth=depth)



"""logging utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

import logging
import os
from typing import Optional

from .console import CONSOLE


LOG_LEVEL: Optional[str] = os.environ.get("LOG_LEVEL", None)
IS_DEBUG: bool = 'DEBUG' in os.environ


class Log:
    """Log class"""

    VERBOSE: int = 5
    DEBUG: int = logging.DEBUG
    WARNING: int = logging.WARNING
    ERROR: int = logging.ERROR

    def __init__(self, name: str):
        logging.basicConfig()
        logging.raiseExceptions = True
        logging.addLevelName(Log.VERBOSE, 'VERBOSE')
        self.logger = logging.getLogger(name)

        if LOG_LEVEL == 'VERBOSE':
            self.level = self.VERBOSE
        elif LOG_LEVEL == 'DEBUG' or IS_DEBUG:
            self.level = self.DEBUG
        elif LOG_LEVEL == 'WARNING':
            self.level = self.WARNING
        else:
            self.level = self.ERROR

    @property
    def level(self) -> int:
        return self.logger.level

    @level.setter
    def level(self, level: int):
        self.logger.setLevel(level)

    def error(self, message: Optional[str] = None, error: Optional[BaseException] = None) -> None:  # noqa
        if self.logger.level <= logging.ERROR:
            self._log(logging.ERROR, message, error)

    def debug(self, message: Optional[str] = None, error: Optional[BaseException] = None) -> None:  # noqa
        if self.logger.level <= logging.DEBUG:
            self._log(logging.DEBUG, message, error)

    def warning(self, message: Optional[str] = None, error: Optional[BaseException] = None) -> None:  # noqa
        if self.logger.level <= logging.WARNING:
            self._log(logging.WARNING, message, error)

    def verbose(self, message: Optional[str] = None, error: Optional[BaseException] = None) -> None:  # noqa
        if self.logger.level <= self.VERBOSE:
            self._log(self.VERBOSE, message, error)

    def _log(self, level: int, message: Optional[str], error: Optional[BaseException] = None) -> None:  # noqa
        if message is not None:
            CONSOLE.stderr(message.strip())
        if error is not None:
            CONSOLE.print_exception()



"""Base Git utility class

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from pathlib import Path
from typing import Optional

# from pygoodle.git.constants import ORIGIN
from pygoodle.git.offline import GitOffline
from pygoodle.git.online import GitOnline

from .repo import Repo


class Submodule(Repo):
    """Class encapsulating base git utilities

    :ivar Path path: Absolute path to repo
    :ivar Path submodule_path: Relative path to submodule
    """

    def __init__(self, repo_path: Path, submodule_path: Path, url: Optional[str] = None, commit: Optional[str] = None,
                 branch: Optional[str] = None, active: Optional[bool] = None):
        """LocalRepo __init__

        :param Path repo_path: Absolute path to repo
        :param Path submodule_path: Relative path to submodule
        :param Optional[str] url: Remote url
        :param Optional[str] commit: Current commit sha stored in git tree
        :param Optional[str] branch: Branch to track
        :param Optional[bool] active: Whether submodule is active
        """

        self.repo_path: Path = repo_path
        self.submodule_path: Path = submodule_path
        super().__init__(repo_path / submodule_path)
        self._url: Optional[str] = url
        self._commit: Optional[str] = commit
        self._branch: Optional[str] = branch
        self._active: Optional[bool] = active

    def __eq__(self, other) -> bool:
        if isinstance(other, Submodule):
            return self.path == other.path
        return False

    def __lt__(self, other: 'Submodule') -> bool:
        return self.path < other.path

    @property
    def exists(self) -> bool:
        is_initialized = GitOffline.is_submodule_initialized(self.repo_path, self.submodule_path)
        is_cloned = GitOffline.is_submodule_cloned(self.repo_path, self.submodule_path)
        return is_initialized and is_cloned

    @property
    def is_initialized(self) -> bool:
        return GitOffline.is_submodule_initialized(self.repo_path, self.submodule_path)

    def absorbgitdirs(self) -> None:
        GitOffline.submodule_absorbgitdirs(self.repo_path, paths=[self.submodule_path])

    def deinit(self, force: bool = False) -> None:
        GitOffline.submodule_deinit(self.repo_path, force=force, paths=[self.submodule_path])

    def init(self) -> None:
        GitOffline.submodule_init(self.repo_path, paths=[self.submodule_path])

    def set_branch(self, branch: str) -> None:
        GitOffline.submodule_set_branch(self.repo_path, self.submodule_path, branch)

    def set_url(self, url: str) -> None:
        GitOffline.submodule_set_url(self.repo_path, self.submodule_path, url)

    def unset_branch(self) -> None:
        GitOffline.submodule_unset_branch(self.repo_path, self.submodule_path)

    def sync(self, recursive: bool = False) -> None:
        GitOffline.submodule_sync(self.repo_path, recursive=recursive, paths=[self.submodule_path])

    def update(self, init: bool = False, depth: Optional[int] = None, single_branch: bool = False,
               jobs: Optional[int] = None, recursive: bool = False, checkout: bool = False,
               rebase: bool = False, merge: bool = False) -> None:
        GitOnline.submodule_update(self.repo_path, init=init, depth=depth, single_branch=single_branch, jobs=jobs,
                                   recursive=recursive, checkout=checkout, merge=merge, rebase=rebase,
                                   paths=[self.submodule_path])



"""unrar utilities

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from typing import Dict, List, Union

from pygoodle.format import Format

from .file_type import FileType


class FileWrapper:

    csv_header: List[str] = [
        'name',
        'size',
        'file_type',
        'downloaded'
    ]

    def __init__(self, name: str, size: int, file_type: Union[FileType, str], downloaded: Union[bool, str] = False):
        self.name: str = name
        self.size: int = int(size)

        if isinstance(file_type, str):
            self.file_type: str = file_type
        elif isinstance(file_type, FileType):
            self.file_type: str = file_type.value
        else:
            raise Exception('Unknown file type')

        if isinstance(downloaded, str):
            self.downloaded: bool = True if downloaded == 'True' else False
        elif isinstance(downloaded, bool):
            self.downloaded: bool = downloaded
        else:
            raise Exception('Unknown downloaded type')

    def __eq__(self, other) -> bool:
        if isinstance(other, FileWrapper):
            return self.name == other.name and self.size == other.size and self.downloaded == other.downloaded
        return False

    def __hash__(self):
        return hash(self.name) ^ hash(self.size) ^ hash(self.downloaded)

    def __lt__(self, other: 'FileWrapper') -> bool:
        return self.name.lower() < other.name.lower()

    @property
    def csv(self) -> List[str]:
        csv = []
        for prop in FileWrapper.csv_header:
            csv.append(getattr(self, prop.lower()))
        return csv

    @property
    def formatted_row(self) -> List[str]:
        return [
            self.name,
            self.get_file_type().formatted_value,
            Format.size(self.size),
            Format.bool(self.downloaded)
        ]

    @property
    def lower_name(self) -> str:
        return self.name.lower()

    def get_file_type(self) -> FileType:
        return FileType(self.file_type)

    @classmethod
    def from_csv(cls, csv: Dict[str, str]) -> 'FileWrapper':
        kwargs = {column: row if row else None for (column, row) in csv.items()}
        return FileWrapper(**kwargs)



"""pygoodle git module __init__

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from .constants import (
    GitConfig,
    HEAD,
    ORIGIN,
    FETCH_URL,
    PUSH_URL,
    GITMODULES
)

from .model.factory import AllBranches
from .model.branch.branch import Branch
from .model.commit import Commit
from .model.branch.local_branch import LocalBranch
from .model.tag.local_tag import LocalTag
from .model.protocol import Protocol
from .model.ref import Ref
from .model.remote import Remote
from .model.branch.remote_branch import RemoteBranch
from .model.tag.remote_tag import RemoteTag
from .model.repo import Repo
from .model.submodule import Submodule
from .model.tag.tag import Tag
from .model.branch.tracking_branch import TrackingBranch

from .offline import GitOffline
from .online import GitOnline



"""file type enum

.. codeauthor:: Joe DeCapo <joe@polka.cat>

"""

from enum import auto, unique

from pygoodle.format import Format
from pygoodle.enum import AutoLowerName


@unique
class FileType(AutoLowerName):
    FILE = auto()
    DIR = auto()

    @property
    def formatted_value(self) -> str:
        if self is FileType.DIR:
            return Format.blue(self.value)
        elif self is FileType.FILE:
            return self.value
        else:
            raise Exception('Unknown file type')
